package ch.unige.aou.service;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.nio.charset.Charset;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.service.GitInfoProperties;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.XMLTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.AouMetadataVersion;
import ch.unige.aou.business.OAISetService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AccessController;
import ch.unige.aou.exception.AouOAIException;
import ch.unige.aou.model.index.PublicationMetadata;
import ch.unige.aou.model.oai.OAISet;
import ch.unige.aou.model.oai.OAIToken;
import ch.unige.aou.model.xml.aou.v1.RepositoryInfo;
import ch.unige.aou.model.xml.oai.v2.DeletedRecordType;
import ch.unige.aou.model.xml.oai.v2.DescriptionType;
import ch.unige.aou.model.xml.oai.v2.GetRecordType;
import ch.unige.aou.model.xml.oai.v2.GranularityType;
import ch.unige.aou.model.xml.oai.v2.HeaderType;
import ch.unige.aou.model.xml.oai.v2.IdentifyType;
import ch.unige.aou.model.xml.oai.v2.ListIdentifiersType;
import ch.unige.aou.model.xml.oai.v2.ListMetadataFormatsType;
import ch.unige.aou.model.xml.oai.v2.ListRecordsType;
import ch.unige.aou.model.xml.oai.v2.ListSetsType;
import ch.unige.aou.model.xml.oai.v2.MetadataFormatType;
import ch.unige.aou.model.xml.oai.v2.MetadataType;
import ch.unige.aou.model.xml.oai.v2.OAIPMHerrorType;
import ch.unige.aou.model.xml.oai.v2.OAIPMHerrorcodeType;
import ch.unige.aou.model.xml.oai.v2.OAIPMHtype;
import ch.unige.aou.model.xml.oai.v2.RecordType;
import ch.unige.aou.model.xml.oai.v2.RequestType;
import ch.unige.aou.model.xml.oai.v2.SetType;
import ch.unige.aou.model.xml.oai.v2.VerbType;
import ch.unige.aou.model.xml.oai.v2.oai_dc.ElementType;
import ch.unige.aou.model.xml.oai.v2.oai_dc.OaiDcType;
import ch.unige.aou.model.xml.oai.v2.oai_dc.ObjectFactory;
import ch.unige.aou.model.xml.oai.v2.oai_identifier.OaiIdentifierType;
import ch.unige.aou.repository.OAISetRepository;
import ch.unige.aou.specification.OAISetSpecification;

@Service
@ConditionalOnBean(AccessController.class)
public class OAIService extends AouService {
  private static final Logger log = LoggerFactory.getLogger(OAIService.class);

  private final ObjectFactory oaiDcFactory = new ObjectFactory();
  private final GitInfoProperties gitInfoProperties;
  private final IdentifierService identifierService;
  private final CommonMetadataService metadataService;
  private JAXBContext jaxbContext;

  private final OAISetService oaiSetService;
  private final OAISetRepository oaiSetRepository;

  private final SearchMgmt searchMgmt;
  private final Map<String, String> xlstList = new HashMap<>();

  private String repositoryLongName;
  private String repositoryEmail;
  private String repositoryProductionDate;
  private String repositoryDomain;
  private String repositoryPrefix;
  private String repositoryName;
  private String repositoryDescription;
  private String repositoryInstitution;
  private String repositoryArchiveHomePage;
  private int oaiPmhPageSize;
  private boolean oaiPmhIsXslt;

  public OAIService(AouProperties aouProperties, MessageService messageService, SearchMgmt searchMgmt,
          OAISetService oaiSetService, GitInfoProperties gitInfoProperties, IdentifierService identifierService,
          CommonMetadataService metadataService, OAISetRepository oaiSetRepository) {
    super(messageService);
    this.searchMgmt = searchMgmt;
    this.oaiSetService = oaiSetService;
    this.oaiSetRepository = oaiSetRepository;
    this.gitInfoProperties = gitInfoProperties;
    this.identifierService = identifierService;
    this.metadataService = metadataService;

    this.oaiPmhIsXslt = aouProperties.getOaiPmh().isXslt();
    this.repositoryLongName = aouProperties.getRepository().getLongName();
    this.repositoryEmail = aouProperties.getRepository().getEmail();
    this.repositoryProductionDate = aouProperties.getRepository().getProductionDate();
    this.repositoryDomain = aouProperties.getRepository().getDomain();
    this.repositoryPrefix = aouProperties.getRepository().getPrefix();
    this.repositoryName = aouProperties.getRepository().getName();
    this.repositoryDescription = aouProperties.getRepository().getDescription();
    this.repositoryInstitution = aouProperties.getRepository().getInstitution();
    this.repositoryArchiveHomePage = aouProperties.getRepository().getArchiveHomePage();
    this.oaiPmhPageSize = aouProperties.getOaiPmh().getPageSize();

    // JAXB contexts
    try {
      this.jaxbContext = JAXBContext.newInstance(OAIPMHtype.class, OaiDcType.class, OaiIdentifierType.class, RepositoryInfo.class);
      // Load XSLT files
      if (this.oaiPmhIsXslt) {
        this.loadXsl();
      }
    } catch (JAXBException | IOException e) {
      throw new SolidifyRuntimeException(e.getMessage(), e);
    }

  }

  public OAIPMHerrorType getError(OAIPMHerrorcodeType code, String message) {
    final OAIPMHerrorType oaiError = new OAIPMHerrorType();
    oaiError.setCode(code);
    oaiError.setValue(message);
    return oaiError;
  }

  public IdentifyType getIdentify(String oaiUrl) {
    final IdentifyType oaiIdentify = new IdentifyType();
    oaiIdentify.setRepositoryName(this.repositoryLongName);
    oaiIdentify.setBaseURL(oaiUrl);
    oaiIdentify.setProtocolVersion("2.0");
    oaiIdentify.getAdminEmail().add(this.repositoryEmail);
    oaiIdentify.setEarliestDatestamp(this.repositoryProductionDate);
    oaiIdentify.setDeletedRecord(DeletedRecordType.PERSISTENT);
    oaiIdentify.setGranularity(GranularityType.YYYY_MM_DD_THH_MM_SS_Z);
    // Add oai-identifier
    final DescriptionType identifierDesc = new DescriptionType();
    final OaiIdentifierType oaiIdentifier = new OaiIdentifierType();
    oaiIdentifier.setScheme(AouConstants.OAI);
    oaiIdentifier.setRepositoryIdentifier(this.repositoryDomain);
    oaiIdentifier.setDelimiter(AouConstants.OAI_SEP);
    oaiIdentifier.setSampleIdentifier(
            AouConstants.OAI + AouConstants.OAI_SEP + this.repositoryDomain + AouConstants.OAI_SEP + this.repositoryPrefix +
                    StringTool.generateResId());
    identifierDesc.setAny(new JAXBElement<>(new QName("http://www.openarchives.org/OAI/2.0/oai-identifier", "oai-identifier"),
            OaiIdentifierType.class, oaiIdentifier));
    oaiIdentify.getDescription().add(identifierDesc);
    // Add local description
    final DescriptionType repoSolution = new DescriptionType();
    final RepositoryInfo repoInfo = new RepositoryInfo();
    repoInfo.setName(this.repositoryName);
    repoInfo.setDescription(this.repositoryDescription);
    repoInfo.setInstitution(this.repositoryInstitution);
    repoInfo.setPoweredBy(AouConstants.AOU_SOLUTION + " v" + this.gitInfoProperties.getBuild().getVersion());
    repoSolution
            .setAny(new JAXBElement<>(new QName("https://archive-ouverte.unige.ch/aou/v1", "repositoryInfo"), RepositoryInfo.class, repoInfo));
    oaiIdentify.getDescription().add(repoSolution);
    return oaiIdentify;
  }

  public GetRecordType getRecord(String baseUrl, String identifier, String metadataPrefix) throws JAXBException {
    // Check Parameters
    this.checkParameters(identifier, metadataPrefix);

    // GetRecordType
    final PublicationMetadata archive = this.metadataService.getMetatadata(this.identifierService.getResIdFromOaiId(identifier));
    if (archive == null) {
      throw new AouOAIException(OAIPMHerrorcodeType.ID_DOES_NOT_EXIST, identifier);
    }
    final GetRecordType getRecord = new GetRecordType();
    getRecord.setRecord(this.getRecord(baseUrl, archive, metadataPrefix));
    return getRecord;
  }

  public RequestType getRequest(String url, String verb, String identifier, String metadataPrefix, String from, String until, String set,
          String token) {
    final RequestType req = new RequestType();
    req.setValue(url);
    if (!StringTool.isNullOrEmpty(verb)) {
      req.setVerb(VerbType.fromValue(verb));
    }
    if (!StringTool.isNullOrEmpty(identifier)) {
      req.setIdentifier(identifier);
    }
    if (!StringTool.isNullOrEmpty(metadataPrefix)) {
      req.setMetadataPrefix(metadataPrefix);
    }
    if (!StringTool.isNullOrEmpty(from)) {
      req.setFrom(from);
    }
    if (!StringTool.isNullOrEmpty(until)) {
      req.setUntil(until);
    }
    if (!StringTool.isNullOrEmpty(set)) {
      req.setSet(set);
    }
    if (!StringTool.isNullOrEmpty(token)) {
      req.setResumptionToken(token);
    }
    return req;
  }

  public VerbType getVerb(String verb) {
    return VerbType.fromValue(verb);
  }

  public VerbType[] getVerbs() {
    return VerbType.values();
  }

  public ListIdentifiersType listIdentifiers(String metadataPrefix, String from, String until, String set, String token) {
    // Check Parameters
    final OAIToken oaiToken = this.checkParameters(metadataPrefix, from, until, set, token);
    // Get Set Query
    String setQuery = null;
    if (oaiToken.getOaiParameters().getSet() != null) {
      setQuery = oaiToken.getOaiParameters().getSet().getQuery();
    }
    final RestCollection<PublicationMetadata> collection = this.searchMgmt
            .search(oaiToken.getOaiParameters().getFrom(), oaiToken.getOaiParameters().getUntil(), setQuery, oaiToken.getPageable());
    if (collection.getPage().getTotalItems() == 0) {
      throw new AouOAIException(OAIPMHerrorcodeType.NO_RECORDS_MATCH, "No identifier found");
    }
    final ListIdentifiersType idList = new ListIdentifiersType();
    final List<PublicationMetadata> list = collection.getData();
    for (final PublicationMetadata archive : list) {
      idList.getHeader().add(this.getIdentifier(archive));
    }
    if (collection.getPage().hasNext()) {
      idList.setResumptionToken(oaiToken.getToken(collection.getPage().getTotalItems(), collection.getPage().getPreviousItems()));
    }
    return idList;
  }

  public ListMetadataFormatsType listMetadataFormats(String identifier) {
    if (identifier != null && !identifier.isEmpty()
            && this.metadataService.getMetatadata(this.identifierService.getResIdFromOaiId(identifier)) == null) {
      throw new AouOAIException(OAIPMHerrorcodeType.ID_DOES_NOT_EXIST, identifier);
    }

    final ListMetadataFormatsType mdFormats = new ListMetadataFormatsType();
    // OAI DC
    final MetadataFormatType oaiDc = new MetadataFormatType();
    oaiDc.setMetadataPrefix(AouConstants.OAI_DC);
    oaiDc.setSchema("http://www.openarchives.org/OAI/2.0/oai_dc.xsd");
    oaiDc.setMetadataNamespace("http://www.openarchives.org/OAI/2.0/oai_dc");
    mdFormats.getMetadataFormat().add(oaiDc);
    // DataCite
    final MetadataFormatType datacite = new MetadataFormatType();
    datacite.setMetadataPrefix(AouConstants.OAI_DATACITE);
    datacite.setSchema("https://schema.datacite.org/meta/kernel-4.3/metadata.xsd");
    datacite.setMetadataNamespace("http://datacite.org/schema/kernel-4");
    mdFormats.getMetadataFormat().add(datacite);

    // schema.org
    final MetadataFormatType schemaorg = new MetadataFormatType();
    schemaorg.setMetadataPrefix(AouConstants.OAI_SCHEMA_ORG);
    schemaorg.setSchema("https://schema.org/Dataset");
    schemaorg.setMetadataNamespace("https://schema.org/Dataset");
    mdFormats.getMetadataFormat().add(schemaorg);
    return mdFormats;
  }

  public ListRecordsType listRecords(String baseUrl, String metadataPrefix, String from, String until, String set, String token)
          throws JAXBException {
    // Check Parameters
    final OAIToken oaiToken = this.checkParameters(metadataPrefix, from, until, set, token);
    String setQuery = null;
    if (oaiToken.getOaiParameters().getSet() != null) {
      setQuery = oaiToken.getOaiParameters().getSet().getQuery();
    }

    final RestCollection<PublicationMetadata> collection = this.searchMgmt
            .search(oaiToken.getOaiParameters().getFrom(), oaiToken.getOaiParameters().getUntil(), setQuery, oaiToken.getPageable());
    if (collection.getPage().getTotalItems() == 0) {
      throw new AouOAIException(OAIPMHerrorcodeType.NO_RECORDS_MATCH, "No record found");
    }
    final ListRecordsType recList = new ListRecordsType();
    final List<PublicationMetadata> list = collection.getData();
    for (final PublicationMetadata archive : list) {
      recList.getRecord().add(this.getRecord(baseUrl, archive, oaiToken.getOaiParameters().getMetadataPrefix()));
    }
    if (collection.getPage().hasNext()) {
      recList.setResumptionToken(oaiToken.getToken(collection.getPage().getTotalItems(), collection.getPage().getPreviousItems()));
    }
    return recList;
  }

  public ListSetsType listSets(String token) {

    /*
     * Pagination may be given by an URL token
     */
    final OAIToken oaiToken = new OAIToken(this.oaiPmhPageSize, Sort.by("name"));
    if (!StringTool.isNullOrEmpty(token)) {
      oaiToken.setToken(token);
    }

    final Page<OAISet> oaiSetsPage = this.getOaiSetPage(oaiToken.getPageable());

    final ListSetsType setTypes = new ListSetsType();
    for (final OAISet oaiSet : oaiSetsPage.getContent()) {

      final SetType setType = new SetType();

      setType.setSetSpec(oaiSet.getSpec());
      setType.setSetName(oaiSet.getName());

      final OaiDcType oaiDc = this.oaiDcFactory.createOaiDcType();
      oaiDc.getTitleOrCreatorOrSubject().add(this.oaiDcFactory.createDescription(this.getOaiDcEntry(oaiSet.getDescription(), null)));
      final DescriptionType descriptionType = new DescriptionType();
      final QName qName = new QName("http://www.openarchives.org/OAI/2.0/oai_dc/", "dc");
      descriptionType.setAny(new JAXBElement<>(qName, OaiDcType.class, oaiDc));

      setType.getSetDescription().add(descriptionType);

      setTypes.getSet().add(setType);
    }

    return setTypes;
  }

  public String printOAIPMH(OAIPMHtype oai, String rootUrl) throws JAXBException {
    final Marshaller m = this.jaxbContext.createMarshaller();
    if (!StringTool.isNullOrEmpty(rootUrl)) {
      m.setProperty("com.sun.xml.bind.xmlHeaders", "<?xml-stylesheet type='text/xsl' href='" + rootUrl + "'?>");
    }
    m.setProperty(AouConstants.XML_NAMESPACE_PREFIX_MAPPER,
            this.metadataService.getAouXmlNamespacePrefixMapper(AouMetadataVersion.getDefaultVersion()));
    m.setProperty(Marshaller.JAXB_SCHEMA_LOCATION, "http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd");
    final JAXBElement<OAIPMHtype> root = this.getXML(oai);
    final ByteArrayOutputStream bos = new ByteArrayOutputStream();
    m.marshal(root, bos);
    return bos.toString();
  }

  private void addOaiSetInfo(HeaderType hdr, String resId) {
    final Pageable pageable = PageRequest.of(0, RestCollectionPage.DEFAULT_SIZE_PAGE);
    final Page<OAISet> oaiSetsPage = this.getOaiSetPage(pageable);
    for (final OAISet oaiSet : oaiSetsPage.getContent()) {
      if (this.searchMgmt.isInSet(resId, oaiSet.getQuery(), pageable)) {
        hdr.getSetSpec().add(oaiSet.getSpec());
      }
    }
  }

  private void checkParameters(String identifier, String metadataPrefix) {
    if (StringTool.isNullOrEmpty(identifier)) {
      throw new AouOAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Identifier is mandatory");
    }
    if (StringTool.isNullOrEmpty(metadataPrefix)) {
      throw new AouOAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Metadata prefix is mandatory");
    }
  }

  private OAIToken checkParameters(String metadataPrefix, String from, String until, String set, String token) {
    final OAIToken oaiToken = new OAIToken(this.oaiPmhPageSize);

    // Token parameter
    if (!StringTool.isNullOrEmpty(token)) {
      oaiToken.setToken(token);
    } else {
      // Metadata Prefix parameter
      if (StringTool.isNullOrEmpty(metadataPrefix)) {
        throw new AouOAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Metadata prefix is mandatory");
      }
      if (!metadataPrefix.equals(AouConstants.OAI_DC) && !metadataPrefix.equals(AouConstants.OAI_DATACITE) && !metadataPrefix
              .equals(AouConstants.OAI_SCHEMA_ORG)) {
        throw new AouOAIException(OAIPMHerrorcodeType.CANNOT_DISSEMINATE_FORMAT, "Metadata prefix not supported");
      }
      oaiToken.getOaiParameters().setMetadataPrefix(metadataPrefix);

      // From parameter
      if (!StringTool.isNullOrEmpty(from)) {
        if (!StringTool.checkDate(from)) {
          throw new AouOAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Wrong 'from' date format (" + from + ")");
        }
        oaiToken.getOaiParameters().setFrom(from);
      }
      // Until parameter
      if (!StringTool.isNullOrEmpty(until)) {
        if (!StringTool.checkDate(until)) {
          throw new AouOAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Wrong 'until' date format (" + until + ")");
        }
        oaiToken.getOaiParameters().setUntil(until);
      }
      if (!StringTool.isNullOrEmpty(from) && !StringTool.isNullOrEmpty(until) && from.length() != until.length()) {
        throw new AouOAIException(OAIPMHerrorcodeType.BAD_ARGUMENT,
                "'from' and 'until' dates do not have the same granularity: " + from + " / " + until);
      }

      // Set parameter
      if (!StringTool.isNullOrEmpty(set)) {

        final OAISet oaiSet = this.oaiSetRepository.findBySpec(set);
        if (oaiSet != null) {
          oaiToken.getOaiParameters().setSet(oaiSet);
        } else {
          throw new AouOAIException(OAIPMHerrorcodeType.NO_SET_HIERARCHY, "Set " + set + " does not exist");
        }
      }
    }
    return oaiToken;
  }

  private String getArchiveUrl() {
    final String archiveUrl = this.repositoryArchiveHomePage;
    if (StringTool.isNullOrEmpty(archiveUrl) || !archiveUrl.endsWith(SolidifyConstants.URL_SEP)) {
      throw new SolidifyRuntimeException("Missing 'aou.repository.archive-home-page' parameter");
    }
    return archiveUrl;
  }

  private String getCreationDate(PublicationMetadata archive) {
    final String creation = (String) archive.getMetadata(AouConstants.CREATION_FIELD.split("\\."));
    final OffsetDateTime creationDate = OffsetDateTime.parse(creation).toInstant().atOffset(ZoneOffset.UTC).withNano(0);
    return creationDate.toString();
  }

  private ElementType getOaiDcEntry(String value, String language) {
    final ElementType fieldValue = new ElementType();
    fieldValue.setValue(value);
    if (!StringTool.isNullOrEmpty(language)) {
      fieldValue.setLang(language);
    }
    return fieldValue;
  }

  private HeaderType getIdentifier(PublicationMetadata archive) {
    final HeaderType hdr = new HeaderType();
    hdr.setIdentifier(this.identifierService.getOaiId(archive.getResId()));
    hdr.setDatestamp(this.getCreationDate(archive));
    this.addOaiSetInfo(hdr, archive.getResId());
    return hdr;
  }

  @SuppressWarnings("squid:S128")
  private Object getMetadata(String baseUrl, String metadataPrefix, PublicationMetadata archive) throws JAXBException {
    String xsl = AouConstants.OAI_DATACITE + "2";
    final String archiveUrl = this.getArchiveUrl();
    switch (metadataPrefix) {
      case AouConstants.OAI_DC:
        xsl += AouConstants.OAI_DC + ".xsl";
        if (this.xlstList.containsKey(xsl)) {
          return this.transformMetadata(archive.getDataCiteMetadata(), xsl, archiveUrl, archive.getResId());
        } else {
          return this.metadataService.getOaiDcMetadata(archive);
        }
      case AouConstants.OAI_SCHEMA_ORG:
        xsl += AouConstants.OAI_SCHEMA_ORG + ".xsl";
        if (this.xlstList.containsKey(xsl)) {
          return this.transformMetadata(archive.getDataCiteMetadata(), xsl, archiveUrl, archive.getResId());
        }
      default:
        throw new AouOAIException(OAIPMHerrorcodeType.NO_METADATA_FORMATS, "No metadata for this record");
    }
  }

  private Page<OAISet> getOaiSetPage(Pageable pageable) {
    /*
     * filter on enabled OAI sets
     */
    final OAISet criteria = new OAISet();
    criteria.setEnabled(true);
    final OAISetSpecification spec = new OAISetSpecification(criteria);
    return this.oaiSetService.findAll(spec, pageable);
  }

  private RecordType getRecord(String baseUrl, PublicationMetadata archive, String metadataPrefix) throws JAXBException {
    final RecordType rec = new RecordType();
    // Header
    rec.setHeader(this.getIdentifier(archive));
    // Metadata
    final MetadataType md = new MetadataType();
    md.setAny(this.getMetadata(baseUrl, metadataPrefix, archive));
    rec.setMetadata(md);
    return rec;
  }

  private JAXBElement<OAIPMHtype> getXML(OAIPMHtype oai) {
    final QName qName = new QName("http://www.openarchives.org/OAI/2.0/", "OAI-PMH");
    return new JAXBElement<>(qName, OAIPMHtype.class, oai);
  }

  private void loadXsl() throws IOException {
    final ResourcePatternResolver resolver = new PathMatchingResourcePatternResolver(this.getClass().getClassLoader());
    for (final org.springframework.core.io.Resource xsl : resolver
            .getResources("classpath*:/" + AouConstants.XSL_HOME + "/" + AouConstants.OAI_DATACITE + "2*.xsl")) {
      log.info("Finding XML transformation: {}", xsl.getFilename());
      this.xlstList.put(xsl.getFilename(), StreamUtils.copyToString(xsl.getInputStream(), Charset.defaultCharset()));
    }

  }

  private Object transformMetadata(String datacite, String xslName, String archiveUrl, String resId) throws JAXBException {
    // XSL parameters
    final Map<String, String> params = new HashMap<>();
    params.put("archiveUrl", archiveUrl);
    params.put("resId", resId);
    // Apply XSL
    final StringReader newMd = new StringReader(XMLTool.transform(datacite, this.xlstList.get(xslName), params));
    return this.jaxbContext.createUnmarshaller().unmarshal(newMd);
  }

}
