package ch.unige.aou.business;

import java.util.List;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;

import ch.unige.aou.controller.AdminController;
import ch.unige.aou.model.notification.EventType;
import ch.unige.aou.model.publication.Comment;
import ch.unige.aou.model.settings.Person;
import ch.unige.aou.repository.CommentRepository;
import ch.unige.aou.specification.CommentSpecification;

@Service
@ConditionalOnBean(AdminController.class)
public class CommentService extends AouResourceService<Comment> {
  private final CommentRepository commentRepository;
  private final PersonService personService;
  private final EventService eventService;

  public CommentService(CommentRepository commentRepository, PersonService personService, @Lazy EventService eventService) {
    this.commentRepository = commentRepository;
    this.personService = personService;
    this.eventService = eventService;
  }

  @Override
  public CommentSpecification getSpecification(Comment resource) {
    return new CommentSpecification(resource);
  }

  public List<Comment> findByPublication(String publicationId) {
    return this.commentRepository.findByPublicationId(publicationId);
  }

  @Override
  public Comment save(Comment comment) {
    final Comment saved = super.save(comment);

    // create an event for a new comment
    if (comment.isOnlyForValidators()) {
      this.eventService.createEvent(comment, EventType.COMMENT_FOR_VALIDATORS_IN_PUBLICATION);
    } else {
      this.eventService.createEvent(comment, EventType.COMMENT_IN_PUBLICATION);
    }

    return saved;
  }

  @Override
  protected void validateItemSpecificRules(Comment comment, BindingResult errors) {
    /**
     * Check that if it is marked as onlyForValidators, that the person linked is validator
     */
    if (comment.isOnlyForValidators()) {
      if (!this.authenticatedUserHasPriviledgedRole() && !this.personService
              .isAllowedToValidate(comment.getCreatorId(), comment.getPublication())) {
        errors.addError(new FieldError(comment.getClass().getSimpleName(), "onlyForValidators", "This value is not permitted."));
      }
    }
  }

  public void changeCommentsPerson(Person creatorToKeep, Person creatorToUpdate) {
    List<Comment> commentList = ((CommentRepository) this.itemRepository).findByPersonResId(creatorToUpdate.getResId());
    for (Comment comment: commentList) {
      comment.setPerson(creatorToKeep);
      this.itemRepository.save(comment);
    }
  }
}
