package ch.unige.aou.business;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.model.PersonWithOrcid;
import ch.unige.solidify.service.PersonWithOrcidService;
import ch.unige.solidify.specification.JoinedTableInValues;

import ch.unige.aou.AouConstants;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AdminController;
import ch.unige.aou.model.display.StructureValidationRight;
import ch.unige.aou.model.notification.NotificationType;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.publication.PublicationSubtype;
import ch.unige.aou.model.security.User;
import ch.unige.aou.model.settings.Person;
import ch.unige.aou.model.settings.Structure;
import ch.unige.aou.model.settings.ValidationRight;
import ch.unige.aou.repository.PersonRepository;
import ch.unige.aou.specification.PersonSpecification;
import ch.unige.aou.specification.PublicationSubtypeSpecification;
import ch.unige.aou.specification.StructureSpecification;
import ch.unige.aou.specification.UserSpecification;

@Service
@ConditionalOnBean(AdminController.class)
public class PersonService extends AouResourceService<Person> implements PersonWithOrcidService {
  static final Logger log = LoggerFactory.getLogger(PersonService.class);

  private final PersonRepository personRepository;
  private final StructureService structureService;
  private final PublicationSubtypeService publicationSubtypeService;
  private final UserService userService;

  public PersonService(PersonRepository personRepository, StructureService structureService,
          PublicationSubtypeService publicationSubtypeService, UserService userService) {
    this.personRepository = personRepository;
    this.structureService = structureService;
    this.publicationSubtypeService = publicationSubtypeService;
    this.userService = userService;
  }

  /************************************/

  @Override
  public PersonSpecification getSpecification(Person resource) {
    return new PersonSpecification(resource);
  }

  /**
   * Return the Person id linked to the Authentication.
   * <p>
   * - It uses REST HTTP requests to get the Person linked to OAuth2 token (call is made on Admin)
   *
   * @param authentication
   * @return String the personId (existing in the Admin module)
   */
  public String getLinkedPersonId(Authentication authentication) {
    return this.authenticatedUserProvider.getPersonId(authentication);
  }

  /**
   * Check from each user linked to the person
   * @return
   */
  public Optional<String> getAuthenticatedUserUnigeCnIndividu() {
    Person person = this.getLinkedPerson((SecurityContextHolder.getContext().getAuthentication()));
    List<User> userList = this.userService.findByPersonResId(person.getResId());
    User userUnige = userList.stream().filter(u-> u.getExternalUid().endsWith(AouConstants.UNIGE_EXTERNAL_UID_SUFFIX)
            || u.getExternalUid().endsWith(AouConstants.TEST_EXTERNAL_UID_SUFFIX)).findFirst().orElse(null);
    return userUnige != null ? this.externalUidToUnigeCnIndividu(userUnige.getExternalUid()): Optional.empty();
  }

  public Person getLinkedPerson(Authentication authentication) {
    String resId = this.getLinkedPersonId(authentication);
    return this.findOne(resId);
  }

  public boolean isAllowedToValidate(String personId, Publication publication) {
    Map<String, List<String>> fullStructuresValidationRights = this.getValidablePublicationSubtypesByStructure(personId);

    List<String> publicationStructureIds = publication.getPublicationStructures().stream()
            .map(publicationStructure -> publicationStructure.getCompositeKey().getStructure().getResId()).collect(Collectors.toList());

    List<String> matchingStructureIds = fullStructuresValidationRights.keySet().stream()
            .filter(structureId -> publicationStructureIds.contains(structureId)).collect(Collectors.toList());

    for (String structureId : matchingStructureIds) {
      List<String> publicationSubtypeIds = fullStructuresValidationRights.get(structureId);
      if (publicationSubtypeIds.contains(publication.getSubtype().getResId())) {
        return true;
      }
    }
    return false;
  }

  public List<Person> getValidators(Publication publication) {

    List<Person> validators = new ArrayList<>();

    // get users with validation rights
    final UserSpecification spec = this.userService.getSpecification(new User());
    spec.setOnlyWithValidationRights(true);
    final List<User> validatorUsers = this.userService.findAll(spec, Pageable.unpaged()).toList();

    // for each user check if they can validate the publication
    for (User validatorUser : validatorUsers) {
      Person validatorPerson = validatorUser.getPerson();
      if (this.isAllowedToValidate(validatorPerson.getResId(), publication) && !validators.contains(validatorPerson)) {
        validators.add(validatorPerson);
      }
    }

    return validators;
  }

  public boolean hasValidationRights(String personId) {
    List<StructureValidationRight> structureValidationRights = this.getStructureValidationRights(personId);
    return !structureValidationRights.isEmpty();
  }

  /**
   * Return a list of StructureValidationRight which is the list of structures the person has validation rights on
   * with their corresponding publication subtypes that the person can validate
   *
   * @param personId
   * @return
   */
  public List<StructureValidationRight> getStructureValidationRights(String personId) {
    StructureSpecification structureSpecification = this.structureService.getSpecification(new Structure());
    /*
     * select a distinct list of structure with validation rights linked to the person
     */
    structureSpecification.addJoinTableInValues(new JoinedTableInValues(ValidationRight.PERSON_RELATION_PROPERTY_NAME,
            ValidationRight.PATH_TO_PERSON + "." + SolidifyConstants.DB_RES_ID, Arrays.asList(personId)));

    final Page<Structure> structuresPage = this.structureService.findAll(structureSpecification, Pageable.unpaged());

    final List<StructureValidationRight> structureValidationRights = new ArrayList<>();
    for (final Structure structure : structuresPage) {
      PublicationSubtypeSpecification publicationSubtypeSpecification = this.publicationSubtypeService
              .getSpecification(new PublicationSubtype());
      publicationSubtypeSpecification.addJoinTableInValues(new JoinedTableInValues(ValidationRight.STRUCTURE_RELATION_PROPERTY_NAME,
              ValidationRight.PATH_TO_PERSON + "." + SolidifyConstants.DB_RES_ID, Arrays.asList(personId)));
      publicationSubtypeSpecification.addJoinTableInValues(new JoinedTableInValues(ValidationRight.STRUCTURE_RELATION_PROPERTY_NAME,
              ValidationRight.PATH_TO_STRUCTURE + "." + SolidifyConstants.DB_RES_ID, Arrays.asList(structure.getResId())));

      List<PublicationSubtype> publicationSubtypes = this.publicationSubtypeService.findAll(publicationSubtypeSpecification, Pageable.unpaged())
              .getContent();

      StructureValidationRight structureValidationRight = new StructureValidationRight(structure);
      structureValidationRight.addPublicationSubtypes(publicationSubtypes);
      structureValidationRights.add(structureValidationRight);
    }

    return structureValidationRights;
  }

  /**
   * Return a map of [structureId ---> List of publication subytpes] representing the person validation rights
   * Child structures are included
   *
   * @param personId
   * @return
   */
  public Map<String, List<String>> getValidablePublicationSubtypesByStructure(String personId) {

    // Get the list of validation rights that have been given to the user (named "parent structures" in the comments below)
    List<StructureValidationRight> structureValidationRights = this.getStructureValidationRights(personId);

    Map<String, List<String>> fullStructuresValidationRights = new HashMap<>();
    for (StructureValidationRight structureValidationRight : structureValidationRights) {
      fullStructuresValidationRights.put(structureValidationRight.getResId(),
              structureValidationRight.getPublicationSubtypes().stream().map(PublicationSubtype::getResId).collect(Collectors.toList()));
    }

    // list of parent structures ids
    List<String> parentStructureIds = fullStructuresValidationRights.keySet().stream().collect(Collectors.toList());

    // Complete the map with all child structures as validation rights are inherited from parent structure
    for (String parentStructureId : parentStructureIds) {

      // get list of descendant structures
      List<String> childStructureIds = this.structureService.getAllChildStructureIds(parentStructureId);

      for (String childStructureId : childStructureIds) {

        if (!fullStructuresValidationRights.containsKey(childStructureId)) {
          // if the child structure is not in the list yet, add it with same publication subtypes ids as parent structure
          fullStructuresValidationRights.put(childStructureId, fullStructuresValidationRights.get(parentStructureId));
        } else {
          // if child structure is already in the list, complete the publication subtypes ids with the new parent structure's ones
          // but do not duplicate them (deduplication is done by using a Set)
          Set<String> set = new HashSet<>();
          set.addAll(fullStructuresValidationRights.get(childStructureId));
          set.addAll(fullStructuresValidationRights.get(parentStructureId));
          List<String> mergedSubtypeIds = set.stream().collect(Collectors.toList());
          fullStructuresValidationRights.put(childStructureId, mergedSubtypeIds);
        }
      }
    }

    return fullStructuresValidationRights;
  }

  @Transactional
  public boolean hasSubscribedToNotificationType(String resId, NotificationType notificationType) {
    Person person = this.findOne(resId);
    return person.getSubscribedNotifications().stream()
            .anyMatch(subscribedNotificationType -> subscribedNotificationType.getResId().equals(notificationType.getResId()));
  }

  public List<Person> findTestPeople() {
    return this.personRepository.findByFirstNameStartingWith(AouConstants.PERMANENT_TEST_DATA_LABEL);
  }

  public Person initDefaultData(AouProperties.Test.User user) {
    final String permanentLabel = AouConstants.PERMANENT_TEST_DATA_LABEL;
    final String firstName = permanentLabel + user.getName() + " First Name";
    final String lastName = permanentLabel + user.getName() + " Last Name";
    Person testPerson = this.personRepository.findByFirstName(firstName);
    if (testPerson == null) {
      testPerson = new Person();
      testPerson.setResId(user.getName());
      testPerson.setFirstName(firstName);
      testPerson.setLastName(lastName);
      this.save(testPerson);
      log.info("Person {} created", testPerson.getFirstName());
    }
    return testPerson;
  }

  @Override
  public PersonWithOrcid getPerson(Authentication authentication) {
    final String authUserId = authentication.getName();
    final User user = this.userService.findByExternalUid(authUserId);
    if (user != null) {
      return user.getPerson();
    } else {
      throw new SolidifyRuntimeException("User referenced by " + authUserId + " is not in database");
    }
  }

  public Person findByOrcid(String orcid) {
    return this.personRepository.findByOrcid(orcid);
  }

  @Override
  public PersonWithOrcid save(PersonWithOrcid person) {
    if (person instanceof Person) {
      return super.save((Person) person);
    } else {
      throw new SolidifyRuntimeException("PersonService cannot save instance of " + person.getClass());
    }
  }
}
