package ch.unige.aou.service;

import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Optional;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.boot.logging.LogLevel;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.FileTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.business.PublicationService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AdminController;
import ch.unige.aou.model.publication.ContributorDTO;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.publication.Publication.PublicationStatus;
import ch.unige.aou.model.tool.CleanTool;
import ch.unige.aou.service.metadata.MetadataExtractor;
import ch.unige.aou.service.storage.StorageService;

@Service
@ConditionalOnBean(AdminController.class)
public class PublicationProcessingService extends AouService {

  private final PublicationService publicationService;
  private final StorageService storageService;
  private final HistoryService historyService;
  private final MetadataService metadataService;

  private final Path nfsSharePath;

  public PublicationProcessingService(MessageService messageService, PublicationService publicationService, StorageService storageService,
          MetadataService metadataService, AouProperties aouProperties, HistoryService historyService) {
    super(messageService);
    this.publicationService = publicationService;
    this.storageService = storageService;
    this.metadataService = metadataService;
    this.historyService = historyService;

    this.nfsSharePath = Paths.get(URI.create(aouProperties.getThesisNfsSharePath()));
    if (!FileTool.checkFile(this.nfsSharePath)) {
      throw new SolidifyRuntimeException("NFS share location (" + this.nfsSharePath + ") does not exist");
    }
  }

  /**
   * Return the Publication saved after the processing
   *
   * @param publicationId Id of the Publication to process
   * @return boolean
   */
  @Transactional
  public Publication processPublication(String publicationId) {

    final Publication publication = this.publicationService.findOne(publicationId);

    this.logPublicationMessage(LogLevel.INFO, publication, "will be processed");

    try {
      switch (publication.getStatus()) {
        case SUBMITTED:
          this.processSubmittedPublication(publication);
          break;
        case CHECKED:
          this.processCheckedPublication(publication);
          break;
        case IN_PREPARATION:
          this.processInPreparationPublication(publication);
          break;
        case CANCEL_EDITION:
          this.revertPublicationChanges(publication);
          break;
        default:
          break;
      }
    } catch (final SolidifyCheckingException e) {
      if (e.getCause() != null) {
        this.inError(publication, e.getCause().getMessage());
      } else {
        this.inError(publication, e.getMessage());
      }
    } catch (final SolidifyProcessingException e) {
      this.inError(publication, e.getMessage());
    } catch (final Exception e) {
      this.logPublicationMessage(LogLevel.ERROR, publication, "Processing error", e);
      this.inError(publication, e.getMessage());
    }

    return this.publicationService.save(publication);
  }

  private void revertPublicationChanges(Publication publication) {
    if (this.publicationService.areThereChangesInDocumentFiles(publication)) {
      // Take the document files from fedora or file system
      this.storageService.revertDocumentFilesAndThumbnailFromStorageInfo(publication);
    }
    publication.setMetadata(publication.getMetadataBackup());
    publication.setMetadataValidationType(Publication.MetadataValidationType.GLOBAL);
    publication.setMetadataBackup(null);
    publication.setStatus(PublicationStatus.COMPLETED);
    publication.setStatusMessage(null);
  }

  private void processSubmittedPublication(Publication publication) throws IOException {
    this.storageService.checkMetadata(publication);
    publication.setStatus(PublicationStatus.CHECKED);
  }

  private void processCheckedPublication(Publication publication) {
    if (this.historyService.hasBeenUpdatingMetadata(publication.getResId())) {
      this.storageService.updateExportMetadata(publication);
    } else {
      this.storageService.generateExportMetadata(publication);
    }
    publication.setStatus(PublicationStatus.IN_PREPARATION);
  }

  private void processInPreparationPublication(Publication publication) throws IOException {
    this.storageService.storePublication(publication);
    publication.setStatus(PublicationStatus.COMPLETED);
    //clean metadataBackup property if needed
    publication.setMetadataBackup(null);
    this.generateDoctorInfoFile(publication);
  }

  /*************************************************************************************************************/

  @Transactional
  public Publication processMetadataToUpgradePublication(String publicationId) {
    Publication publication = this.publicationService.findOne(publicationId);

    // metadata is automatically upgraded after a find --> we only need to save the Publication to persist the upgrade
    // note: if metadata have already been updated, the field is not dirty here and we don't need to save the publication again.
    if (publication.isMetadataDirty()) {
      this.logPublicationMessage(LogLevel.INFO, publication, "will be processed for metadata version upgrade");
      publication = this.publicationService.simpleSave(publication);
    }
    return publication;
  }

  /*************************************************************************************************************/

  private void logPublicationMessage(LogLevel logLevel, Publication publication, String message, Exception... e) {
    this.logObjectWithStatusMessage(logLevel, publication.getClass().getSimpleName(), publication.getResId(),
            publication.getStatus().toString(), publication.getTitle(), message, e);
  }

  private void inError(Publication publication, String message) {
    publication.setErrorStatusWithMessage(message);
    this.logPublicationMessage(LogLevel.ERROR, publication, message);
  }

  //Generate info txt file
  public void generateDoctorInfoFile(Publication publication) throws IOException {
    // For Thesis create a file on the share drive containing doctor's address
    if (publication.getSubtype().getResId().equals(AouConstants.DEPOSIT_SUBTYPE_THESE_ID)) {
      // Check if there is already an author
      final MetadataExtractor metadataExtractor = this.metadataService.getMetadataExtractor(publication.getMetadataVersion());
      final Object depositDoc = metadataExtractor.createDepositDocObjectFromXml(publication.getMetadata());
      String thesisRef = metadataExtractor.getLocalNumber(depositDoc);
      thesisRef = CleanTool.cleanHyphens(thesisRef, true).replaceAll("\\s", "").replaceAll("\\/", "");
      String doctorEmail = metadataExtractor.getDoctorEmail(depositDoc);
      String doctorAddress = metadataExtractor.getDoctorAddress(depositDoc);
      final Optional<ContributorDTO> authorOpt = metadataExtractor.getFirstAuthor(depositDoc);

      final String fileName = authorOpt.get().getLastName() + "_" + authorOpt.get().getFirstName() + "_" + thesisRef + AouConstants.TXT_EXT;
      Files.createFile(this.nfsSharePath.resolve(fileName));
      //write data to the file
      try (BufferedWriter bufferedWriter = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(this.nfsSharePath.resolve(fileName).toFile()), StandardCharsets.UTF_8))) {
        bufferedWriter.write("Email du/de la doctorant-e");
        bufferedWriter.newLine();
        bufferedWriter.write("  " + doctorEmail);
        bufferedWriter.newLine();
        bufferedWriter.write("Adresse postale du/de la doctorant-e");
        bufferedWriter.newLine();
        bufferedWriter.write("  " + doctorAddress.replaceAll("\\r|\\n", " ")); // remove all break lines
      }
    }
  }
}
