package ch.unige.aou.service;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.business.PublicationService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AdminController;
import ch.unige.aou.message.PublicationFromFedoraMessage;

@Service
@ConditionalOnBean(AdminController.class)
public class SolrService extends AouService {
  private static final Logger log = LoggerFactory.getLogger(SolrService.class);

  private String solrUrl;
  private final PublicationService publicationService;

  private static final String TOTAL_ITEMS = "numFound";
  private static final String PID = "pid";
  private static final String DOCS = "docs";

  public SolrService(AouProperties aouProperties, MessageService messageService, PublicationService publicationService) {
    super(messageService);
    this.publicationService = publicationService;
    this.solrUrl = aouProperties.getStorage().getSolrUrl();
  }

  public boolean startPublicationsImportFromFedora(String personId, String solrConditions) {
    //Call to solr api
    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    RestTemplate client = restTemplateBuilder.build();

    String solrUrlWithConditions = this.solrUrl.replace("{solrConditions}", solrConditions);

    final String jsonString = client.getForObject(solrUrlWithConditions, String.class);
    JSONObject jsonObject = new JSONObject(jsonString).getJSONObject("response");

    int totalItems = jsonObject.getInt(TOTAL_ITEMS);
    if (totalItems <= 0) {
      log.info("There is no publication from solr api corresponding to '{}' condition", solrConditions);
      return false;
    } else {
      log.info("{} objects will be verified to check if they need to be imported from Fedora}", totalItems);
    }

    //Get each pid from the response and create a jms message
    JSONArray pidsArray = this.getArrayOrNull(jsonObject, DOCS);
    List<String> pidsFromFedora = new ArrayList<>();

    if (pidsArray != null && !pidsArray.isEmpty()) {
      pidsFromFedora = StreamSupport.stream(pidsArray.spliterator(), false)
              .map(a -> (JSONObject) a)
              .map(a -> a.getString(PID))
              .collect(Collectors.toList());
    }
    pidsFromFedora.forEach(pid -> {
      if (!this.publicationService.existsByArchiveId(pid)) {
        SolidifyEventPublisher.getPublisher().publishEvent(new PublicationFromFedoraMessage(pid, personId));
      } else {
        log.info("pid {} already exists in database", pid);
      }
    });

    return true;
  }

  protected JSONArray getArrayOrNull(JSONObject jsonObject, String propertyName) {
    try {
      if (jsonObject != null && jsonObject.has(propertyName) && !jsonObject.isNull(propertyName)) {
        return jsonObject.getJSONArray(propertyName);
      }
    } catch (JSONException e) {
      return null;
    }
    return null;
  }
}
