package ch.unige.aou.service.admin;

import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.springframework.core.env.Environment;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.SolidifyClientApplication;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.service.ResourceClientService;
import ch.unige.solidify.util.AbstractRestClientTool;
import ch.unige.solidify.util.CollectionTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.model.display.StructureWithJoinInfosDTO;
import ch.unige.aou.model.publication.Contributor;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.rest.PublicationImportDTO;
import ch.unige.aou.model.settings.ResearchGroup;
import ch.unige.aou.model.settings.Structure;
import ch.unige.aou.rest.AouActionName;
import ch.unige.aou.rest.ModuleName;
import ch.unige.aou.rest.ResourceName;

@Service
public class PublicationClientService extends ResourceClientService<Publication> implements PublicationClientServiceInterface {

  public PublicationClientService(AbstractRestClientTool restClientTool, Environment env) {
    super(SolidifyClientApplication.getApplicationPrefix(), ModuleName.ADMIN, SolidifyConstants.URL_SEP + ResourceName.PUBLICATIONS,
            Publication.class, restClientTool, env);
  }

  @Override
  public List<Publication> findAll() {
    return this.findAll(Publication.class);
  }

  @Override
  public List<Publication> searchByProperties(Map<String, String> properties) {
    return this.searchByProperties(properties, Publication.class);
  }

  @Override
  public List<ResearchGroup> getResearchGroups(String publicationId) {
    return this.findAllLinkedResources(publicationId, ResourceName.RESEARCH_GROUPS, ResearchGroup.class);
  }

  @Override
  public void addResearchGroup(String publicationId, String researchGroupId) {
    this.addItem(publicationId, ResourceName.RESEARCH_GROUPS, researchGroupId, ResearchGroup.class);
  }

  @Override
  public void removeResearchGroup(String publicationId, String researchGroupId) {
    this.removeItem(publicationId, ResourceName.RESEARCH_GROUPS, researchGroupId);
  }

  @Override
  public List<StructureWithJoinInfosDTO> getStructures(String publicationId) {
    return this.findAllLinkedResources(publicationId, ResourceName.STRUCTURES, StructureWithJoinInfosDTO.class);
  }

  @Override
  public void addStructure(String publicationId, String structureId) {
    this.addItem(publicationId, ResourceName.STRUCTURES, structureId, Structure.class);
  }

  @Override
  public void removeStructure(String publicationId, String structureId) {
    this.removeItem(publicationId, ResourceName.STRUCTURES, structureId);
  }

  @Override
  public Publication findByTitle(String title) {
    final Map<String, String> properties = new HashMap<>();
    properties.put("title", title);
    List<Publication> publications = this.searchByProperties(properties);
    if (publications.size() == 1) {
      return publications.get(0);
    } else if (publications.size() > 1) {
      throw new SolidifyRuntimeException("Too many publication with title '" + title + "' found");
    } else {
      throw new SolidifyRuntimeException("No publication with title '" + title + "' found");
    }
  }

  @Override
  public List<Publication> listMyPublications(Map<String, String> params) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + AouActionName.LIST_MY_PUBLICATIONS
            + SIZE_PARAM + RestCollectionPage.MAX_SIZE_PAGE;
    return this.listPublications(url, params);
  }

  @Override
  public List<Publication> listValidablePublications(Map<String, String> params) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + AouActionName.LIST_VALIDABLE_PUBLICATIONS
            + SIZE_PARAM + RestCollectionPage.MAX_SIZE_PAGE;
    return this.listPublications(url, params);
  }

  @Override
  public List<Contributor> getContributors(String publicationId) {
    return this.findAllLinkedResources(publicationId, ResourceName.CONTRIBUTORS, Contributor.class);
  }

  private List<Publication> listPublications(String url, Map<String, String> params) {
    StringBuilder urlBuilder = new StringBuilder(url);
    for (final Map.Entry<String, String> entry : params.entrySet()) {
      urlBuilder.append("&")
              .append(entry.getKey())
              .append("=")
              .append(entry.getValue());
    }
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final String jsonString = restTemplate.getForObject(urlBuilder.toString(), String.class);
    return CollectionTool.getList(jsonString, Publication.class);
  }

  @Override
  public Result submit(String publicationId) {
    return this.postAction(publicationId, AouActionName.SUBMIT);
  }

  @Override
  public Result submitForApproval(String publicationId) {
    return this.postAction(publicationId, AouActionName.SUBMIT_FOR_VALIDATION);
  }

  @Override
  public Result submitForApproval(String publicationId, String validationStructureId) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + publicationId
            + SolidifyConstants.URL_SEP + AouActionName.SUBMIT_FOR_VALIDATION;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    HttpEntity<String[]> requestEntity = new HttpEntity<>(new String[] { validationStructureId });
    return restTemplate.postForObject(url, requestEntity, Result.class);
  }

  @Override
  public Result askFeedback(String publicationId, String reason) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + publicationId
            + SolidifyConstants.URL_SEP + AouActionName.ASK_FEEDBACK + "?reason={reason}";
    Map<String, String> params = new HashMap<>();
    params.put("reason", reason);
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.postForObject(url, null, Result.class, params);
  }

  @Override
  public Result updateStatusToDelete(String publicationId, String reason) {
    return this.postAction(publicationId, AouActionName.DELETE);
  }

  @Override
  public Result resume(String publicationId) {
    return this.postAction(publicationId, AouActionName.RESUME);
  }

  @Override
  public Result updatePublicationListWithResearchGroup(String researchGroupId, String[] publicationIds) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + AouActionName.UPDATE_PUBLICATIONS_RESEARCH_GROUP
            + "?" + AouConstants.DB_RESEARCH_GROUP_ID + "=" + researchGroupId;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.patchForObject(url, new HttpEntity<>(publicationIds), Result.class);
  }

  @Override
  public Result updatePublicationListWithStructure(String structureId, String[] publicationIds) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + AouActionName.UPDATE_PUBLICATIONS_STRUCTURE
            + "?" + AouConstants.DB_STRUCTURE_ID + "=" + structureId;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.patchForObject(url, new HttpEntity<>(publicationIds), Result.class);
  }

  @Override
  public Publication createFromDOI(String doi) {
    PublicationImportDTO publicationImportDTO = new PublicationImportDTO();
    publicationImportDTO.setDoi(doi);
    return this.createFromIdentifier(publicationImportDTO);
  }

  @Override
  public Publication createFromDOI(String doi, String creatorId) {
    PublicationImportDTO publicationImportDTO = new PublicationImportDTO();
    publicationImportDTO.setDoi(doi);
    publicationImportDTO.setCreatorId(creatorId);
    return this.createFromIdentifier(publicationImportDTO);
  }

  @Override
  public Publication createFromPMID(String pmid) {
    PublicationImportDTO publicationImportDTO = new PublicationImportDTO();
    publicationImportDTO.setPmid(pmid);
    return this.createFromIdentifier(publicationImportDTO);
  }

  private Publication createFromIdentifier(PublicationImportDTO publicationImportDTO) {
    String url = this.getResourceUrl() + SolidifyConstants.URL_SEP + AouActionName.CREATE_FROM_IDENTIFIER;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.postForObject(url, publicationImportDTO, this.getResourceClass());
  }

  @Override
  public String validateMetadata(String publicationId) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + publicationId
            + SolidifyConstants.URL_SEP + AouActionName.VALIDATE_METADATA;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.postForObject(url, HttpMethod.POST, String.class);
  }

  @Override
  public Publication uploadThumbnail(String publicationId, Resource resource) {
    return this.uploadImage(publicationId, AouActionName.UPLOAD_THUMBNAIL, resource);
  }

  @Override
  public void downloadThumbnail(String publicationId, Path path) {
    this.downloadImage(publicationId, AouActionName.DOWNLOAD_THUMBNAIL, path);
  }

  @Override
  public void deleteThumbnail(String publicationId) {

    this.deleteAction(publicationId, AouActionName.DELETE_THUMBNAIL);
  }

  @Override
  public void unlinkContributor(String publicationId) {
    String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + publicationId
            + SolidifyConstants.URL_SEP + AouActionName.UNLINK_CONTRIBUTOR;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    restTemplate.postForObject(url, null, Result.class);
  }

  @Override
  public Publication clonePublication(String publicationId) {
    final String url = this.getResourceUrl()
            + SolidifyConstants.URL_SEP + publicationId
            + SolidifyConstants.URL_SEP + AouActionName.CLONE;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.postForObject(url, null, Publication.class);
  }
}
