package ch.unige.aou.controlleradvice;

import static ch.unige.solidify.controller.SolidifyControllerAdvice.StackTraceDisplay.WITH_STACK_TRACE;

import javax.servlet.http.HttpServletRequest;

import org.apache.catalina.connector.ClientAbortException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.servlet.error.ErrorAttributes;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;

import ch.unige.solidify.controller.SolidifyControllerAdvice;
import ch.unige.solidify.repository.SolidifyError;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.exception.AoUMetadataParseException;
import ch.unige.aou.exception.AouError;
import ch.unige.aou.exception.AouMetadataValidationException;
import ch.unige.aou.exception.AouRuntimeException;

@ControllerAdvice
public class AoUControllerAdvice extends SolidifyControllerAdvice {
  private static final Logger log = LoggerFactory.getLogger(AoUControllerAdvice.class);

  private final ErrorAttributes errorAttributes;
  private final HttpHeaders headers = new HttpHeaders();

  public AoUControllerAdvice(ErrorAttributes errorAttributes, MessageService messageService) {
    super(errorAttributes, messageService);
    this.errorAttributes = errorAttributes;
    this.headers.add("Content-Type", "application/hal+json");
  }

  // ClientAbortException belongs to Tomcat, handle it at project level since Solidify-Backend is container agnostic
  @ExceptionHandler(ClientAbortException.class)
  public void handleClientAbortException(ClientAbortException exception, HttpServletRequest request) {
    final String message = "ClientAbortException generated by request {} {} from remote address {} with X-FORWARDED-FOR {}";
    final String headerXFF = request.getHeader("X-FORWARDED-FOR");
    log.warn(message, request.getMethod(), request.getRequestURL(), request.getRemoteAddr(), headerXFF);
  }

  @ExceptionHandler(AoUMetadataParseException.class)
  public ResponseEntity<SolidifyError> handleException(AoUMetadataParseException exception, HttpServletRequest request) {
    String message = "";
    if (exception.getCause() != null) {
      message = exception.getCause().getMessage();
    }
    return this.buildResponse(exception, request, message, HttpStatus.BAD_REQUEST, WITH_STACK_TRACE);
  }

  @ExceptionHandler(AouMetadataValidationException.class)
  public ResponseEntity<SolidifyError> handleException(AouMetadataValidationException exception, HttpServletRequest request) {

    final SolidifyError solidifyError = new SolidifyError(this.errorAttributes, request);
    solidifyError.setValidationError(exception.getValidationError());
    solidifyError.setMessage(exception.getMessage());
    solidifyError.setStatus(HttpStatus.BAD_REQUEST);

    this.logException(exception, request, StackTraceDisplay.WITHOUT_STACK_TRACE);
    return ResponseEntity.badRequest().headers(this.headers).body(solidifyError);
  }

  @ExceptionHandler(AouRuntimeException.class)
  public ResponseEntity<AouError> handleException(AouRuntimeException exception, HttpServletRequest request) {
    AouError aouError = exception.getAouError();
    aouError.setPath(request.getRequestURI());

    this.logException(exception, request, WITH_STACK_TRACE);

    aouError.setMessage(exception.getMessage());
    aouError.setError(exception.getMessage());
    aouError.setStatus(HttpStatus.BAD_REQUEST);

    return (ResponseEntity.status(HttpStatus.BAD_REQUEST).headers(this.headers)).body(aouError);
  }
}
