package ch.unige.aou.service;

import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.index.indexing.IndexingService;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.DataMgmtController;
import ch.unige.aou.model.index.PublicationIndexEntry;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.service.rest.trusted.TrustedPublicationRemoteResourceService;

@Service
@ConditionalOnBean(DataMgmtController.class)
public class IndexingProcessingService extends AouService {

  private static final Logger log = LoggerFactory.getLogger(IndexingProcessingService.class);

  private final String publishedPublicationsIndexName;
  private final String inProgressPublicationsIndexName;

  private final CommonMetadataService commonMetadataService;
  private final TrustedPublicationRemoteResourceService remotePublicationService;

  private final IndexingService<PublicationIndexEntry> publicationIndexingService;

  public IndexingProcessingService(AouProperties aouProperties, MessageService messageService, CommonMetadataService commonMetadataService,
          TrustedPublicationRemoteResourceService remotePublicationService, IndexingService<PublicationIndexEntry> publicationIndexingService) {
    super(messageService);
    this.remotePublicationService = remotePublicationService;
    this.publicationIndexingService = publicationIndexingService;
    this.commonMetadataService = commonMetadataService;

    this.publishedPublicationsIndexName = aouProperties.getIndexing().getPublishedPublicationsIndexName();
    this.inProgressPublicationsIndexName = aouProperties.getIndexing().getInProgressPublicationsIndexName();
  }

  @Transactional
  public void indexPublication(String publicationId) {
    try {
      Publication publication = this.remotePublicationService.findOne(publicationId);
      PublicationIndexEntry publicationIndexEntry = this.commonMetadataService.getPublicationIndexEntry(publication);

      // always store in index used on admin
      this.storeInIndex(this.inProgressPublicationsIndexName, publicationIndexEntry, publication);

      // if COMPLETED, store in index used on public interface as well
      if (publication.getStatus() == Publication.PublicationStatus.COMPLETED) {
        this.storeInIndex(this.publishedPublicationsIndexName, publicationIndexEntry, publication);
      }

    } catch (SolidifyResourceNotFoundException e) {
      // publication has been deleted
      this.deleteFromIndexIfExists(this.inProgressPublicationsIndexName, publicationId);
      this.deleteFromIndexIfExists(this.publishedPublicationsIndexName, publicationId);
    } catch (Exception e) {
      log.error("Unable to index object {}", publicationId, e);
    }
  }

  private void storeInIndex(String indexName, PublicationIndexEntry publicationIndexEntry, Publication publication) {
    publicationIndexEntry.setIndex(indexName);
    this.deleteFromIndexIfExists(indexName, publicationIndexEntry.getResId());
    this.publicationIndexingService.save(indexName, publicationIndexEntry);
    log.info("Publication '{}' ({}) indexed in '{}' index", publication.getResId(), StringTool.truncateWithEllipsis(publication.getTitle(), 20),
            indexName);
  }

  private void deleteFromIndexIfExists(String indexName, String publicationId) {
    if (this.publicationIndexingService.findOne(indexName, publicationId) != null) {
      this.publicationIndexingService.deleteList(indexName, List.of(publicationId));
    }
  }
}
