package ch.unige.aou.specification;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.springframework.data.jpa.domain.Specification;

import ch.unige.solidify.util.StringTool;

import ch.unige.aou.model.publication.Contributor;
import ch.unige.aou.model.search.ContributorSearchType;

public class ContributorSearchSpecification implements Specification<Contributor> {
  private static final long serialVersionUID = -6039373889994880237L;

  private final String searchTerm;
  private final ContributorSearchType searchType;

  public ContributorSearchSpecification(String searchTerm, ContributorSearchType searchType) {
    this.searchTerm = searchTerm;
    this.searchType = searchType;
  }

  @Override
  public Predicate toPredicate(Root<Contributor> root, CriteriaQuery<?> query, CriteriaBuilder builder) {

    final List<Predicate> predicatesList = new ArrayList<>();

    if (!this.searchTerm.contains(",")) {
      this.fillLooseNamesConditions(root, builder, predicatesList);
    } else {
      this.fillFirstNameAndLastNameConditions(root, builder, predicatesList);
    }

    Predicate cnIndividuPredicate = this.getCnIndividuPredicate(root, builder);
    if (cnIndividuPredicate != null) {
      predicatesList.add(cnIndividuPredicate);
    }

    return builder.and(predicatesList.toArray(new Predicate[predicatesList.size()]));
  }

  private void fillLooseNamesConditions(Root<Contributor> root, CriteriaBuilder builder, List<Predicate> predicatesList) {
    List<String> termsList = Arrays.stream(this.searchTerm.split(" ")).filter(term -> !StringTool.isNullOrEmpty(term))
            .collect(Collectors.toList());

    Predicate orNamesPredicate = builder.or();
    for (String term : termsList) {
      term = term.trim();
      if (!StringTool.isNullOrEmpty(term)) {
        Predicate likeFirstName = builder.like(root.get("firstName"), "%" + term + "%");
        orNamesPredicate.getExpressions().add(likeFirstName);
        Predicate likeLastName = builder.like(root.get("lastName"), "%" + term + "%");
        orNamesPredicate.getExpressions().add(likeLastName);
      }
    }
    if (!orNamesPredicate.getExpressions().isEmpty()) {
      predicatesList.add(orNamesPredicate);
    }
  }

  private void fillFirstNameAndLastNameConditions(Root<Contributor> root, CriteriaBuilder builder, List<Predicate> predicatesList) {
    String[] names = this.searchTerm.split(",");
    if (names.length > 0 && names.length <= 2) {

      String lastName = names[0].trim();
      Predicate likeLastName = builder.like(root.get("lastName"), "%" + lastName + "%");
      predicatesList.add(likeLastName);

      if (names.length == 2) {
        String firstName = names[1].trim();
        Predicate likeFirstName = builder.like(root.get("firstName"), "%" + firstName + "%");
        predicatesList.add(likeFirstName);
      }
    }
  }

  private Predicate getCnIndividuPredicate(Root<Contributor> root, CriteriaBuilder builder) {
    if (this.searchType == ContributorSearchType.UNIGE) {
      return builder.isNotNull(root.get("cnIndividu"));
    } else if (this.searchType == ContributorSearchType.NON_UNIGE) {
      return builder.isNull(root.get("cnIndividu"));
    }
    return null;
  }
}
