package ch.unige.aou.service;

import static ch.unige.aou.AouConstants.DOCUMENT_FILE_TYPE_AGREEMENT_VALUE;
import static ch.unige.aou.AouConstants.DOCUMENT_FILE_TYPE_PUBLICATION_MODE_VALUE;

import java.io.IOException;
import java.io.StringReader;
import java.time.LocalDate;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.format.DateTimeParseException;
import java.time.temporal.ChronoField;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import javax.xml.bind.JAXB;
import javax.xml.parsers.ParserConfigurationException;

import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import org.springframework.stereotype.Service;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.XMLTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.AouMetadataVersion;
import ch.unige.aou.AouXmlNamespacePrefixMapper;
import ch.unige.aou.AouXmlNamespacePrefixMapperV1;
import ch.unige.aou.AouXmlNamespacePrefixMapperV2_0;
import ch.unige.aou.AouXmlNamespacePrefixMapperV2_1;
import ch.unige.aou.AouXmlNamespacePrefixMapperV2_2;
import ch.unige.aou.AouXmlNamespacePrefixMapperV2_3;
import ch.unige.aou.AouXmlNamespacePrefixMapperV2_4;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.exception.AoUMetadataParseException;
import ch.unige.aou.model.StatusHistory;
import ch.unige.aou.model.index.PublicationIndexEntry;
import ch.unige.aou.model.index.PublicationMetadata;
import ch.unige.aou.model.publication.AbstractContributor;
import ch.unige.aou.model.publication.CollaborationDTO;
import ch.unige.aou.model.publication.ContributorDTO;
import ch.unige.aou.model.publication.DateType;
import ch.unige.aou.model.publication.DocumentFile;
import ch.unige.aou.model.publication.DocumentFileType;
import ch.unige.aou.model.publication.MetadataClassification;
import ch.unige.aou.model.publication.MetadataCollection;
import ch.unige.aou.model.publication.MetadataCorrection;
import ch.unige.aou.model.publication.MetadataDates;
import ch.unige.aou.model.publication.MetadataEmbargo;
import ch.unige.aou.model.publication.MetadataFile;
import ch.unige.aou.model.publication.MetadataFunding;
import ch.unige.aou.model.publication.MetadataLink;
import ch.unige.aou.model.publication.MetadataTextLang;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.publication.PublicationSubtypeDTO;
import ch.unige.aou.model.publication.PublicationSubtypeDocumentFileTypeDTO;
import ch.unige.aou.model.settings.ResearchGroup;
import ch.unige.aou.model.settings.Structure;
import ch.unige.aou.model.xml.deposit.v2_4.DepositDoc;
import ch.unige.aou.service.metadata.DepositDocAdapter;
import ch.unige.aou.service.metadata.DepositDocV1Adapter;
import ch.unige.aou.service.metadata.DepositDocV2Adapter;
import ch.unige.aou.service.rest.trusted.TrustedDocumentFileRemoteResourceService;
import ch.unige.aou.service.rest.trusted.TrustedDocumentFileTypeRemoteResourceService;
import ch.unige.aou.service.rest.trusted.TrustedPublicationRemoteResourceService;
import ch.unige.aou.service.rest.trusted.TrustedPublicationSubtypeRemoteResourceService;
import ch.unige.aou.service.rest.trusted.TrustedStructureRemoteResourceService;

@Service
public class CommonMetadataService extends AouService {

  private static final String INDEX_RES_ID_FIELD = SolidifyConstants.RES_ID_FIELD;
  private static final String INDEX_ACRONYM_FIELD = "acronym";
  private static final String METADATA_ERROR_VERSION = "metadata.error.version";

  private final TrustedStructureRemoteResourceService remoteStructureService;
  private final TrustedPublicationRemoteResourceService remotePublicationService;
  private final TrustedPublicationSubtypeRemoteResourceService remotePublicationSubtypeService;
  private final TrustedDocumentFileRemoteResourceService remoteDocumentFileService;
  private final TrustedDocumentFileTypeRemoteResourceService remoteDocumentFileTypeService;

  private final String[] metadataDateFormat;

  public CommonMetadataService(AouProperties aouProperties, MessageService messageService,
          TrustedStructureRemoteResourceService remoteStructureService,
          TrustedPublicationRemoteResourceService remotePublicationService,
          TrustedDocumentFileRemoteResourceService remoteDocumentFileService,
          TrustedPublicationSubtypeRemoteResourceService remotePublicationSubtypeService,
          TrustedDocumentFileTypeRemoteResourceService remoteDocumentFileTypeService) {

    super(messageService);

    this.remoteStructureService = remoteStructureService;
    this.remotePublicationService = remotePublicationService;
    this.remoteDocumentFileService = remoteDocumentFileService;
    this.remotePublicationSubtypeService = remotePublicationSubtypeService;
    this.remoteDocumentFileTypeService = remoteDocumentFileTypeService;
    this.metadataDateFormat = aouProperties.getParameters().getMetadataDateFormat();
  }

  public AouXmlNamespacePrefixMapper getAouXmlNamespacePrefixMapper(AouMetadataVersion version) {
    switch (version) {
      case V1_0:
        return new AouXmlNamespacePrefixMapperV1();
      case V2_0:
        return new AouXmlNamespacePrefixMapperV2_0();
      case V2_1:
        return new AouXmlNamespacePrefixMapperV2_1();
      case V2_2:
        return new AouXmlNamespacePrefixMapperV2_2();
      case V2_3:
        return new AouXmlNamespacePrefixMapperV2_3();
      case V2_4:
        return new AouXmlNamespacePrefixMapperV2_4();
      default:
        throw new SolidifyProcessingException(this.messageService.get(METADATA_ERROR_VERSION, new Object[] { version }));
    }
  }

  public AouMetadataVersion detectVersionFromXmlMetadata(String xmlMetadata) {
    try {
      Document doc = XMLTool.parseXML(xmlMetadata);
      String namespace = doc.getNamespaceURI();

      for (AouMetadataVersion version : AouMetadataVersion.getAllVersions()) {
        AouXmlNamespacePrefixMapper mapper = this.getAouXmlNamespacePrefixMapper(version);
        if (XMLTool.getFirstNode(doc, "/" + AouConstants.XML_NAMESPACE_DEPOSIT_PREFIX + ":deposit_doc", mapper.getNamespacePrefixes()) != null) {
          return version;
        }
      }

      throw new SolidifyProcessingException("namespace '" + namespace + "' does not correspond to any metadata version");
    } catch (SAXException e) {
      throw new AoUMetadataParseException("Unable to parse metadata", e);
    } catch (ParserConfigurationException | IOException e) {
      throw new SolidifyRuntimeException("unable to detect metadata version", e);
    }
  }

  public DepositDocAdapter getDepositDocAdapter(Publication publication) {
    return this.getDepositDocAdapter(this.detectVersionFromXmlMetadata(publication.getMetadata()));
  }

  public Object getDepositDoc(Publication publication) {
    AouMetadataVersion version = this.detectVersionFromXmlMetadata(publication.getMetadata());
    switch (version) {
      case V1_0:
        return JAXB.unmarshal(new StringReader(publication.getMetadata()), ch.unige.aou.model.xml.deposit.v1.DepositDoc.class);
      case V2_0:
      case V2_1:
      case V2_2:
      case V2_3:
      case V2_4:
        return JAXB.unmarshal(new StringReader(publication.getMetadata()), DepositDoc.class);
      default:
        throw new SolidifyProcessingException(this.messageService.get(METADATA_ERROR_VERSION, new Object[] { version }));
    }
  }

  public DepositDocAdapter getDepositDocAdapter(AouMetadataVersion version) {
    switch (version) {
      case V1_0:
        return new DepositDocV1Adapter();
      case V2_0:
      case V2_1:
      case V2_2:
      case V2_3:
      case V2_4:
        return new DepositDocV2Adapter();
      default:
        throw new SolidifyProcessingException(this.messageService.get(METADATA_ERROR_VERSION, new Object[] { version }));
    }
  }

  /*****************************************************************************************************/

  public PublicationIndexEntry getPublicationIndexEntry(Publication publication) {
    PublicationIndexEntry publicationIndexEntry = new PublicationIndexEntry();
    publicationIndexEntry.setResId(publication.getResId());
    publicationIndexEntry.setMetadata(this.xmlMetadataToMap(publication));

    // add publication status

    // add publication creator
    Map<String, String> creatorMap = new HashMap<>();
    creatorMap.put(INDEX_RES_ID_FIELD, publication.getCreator().getResId());
    creatorMap.put("firstName", publication.getCreator().getFirstName());
    creatorMap.put("lastName", publication.getCreator().getLastName());
    creatorMap.put("fullName", publication.getCreator().getFullName());
    publicationIndexEntry.getMetadata().put("creator", creatorMap);

    // add publication technical infos
    Map<String, Object> technicalInfosMap = new HashMap<>();
    technicalInfosMap.put("status", publication.getStatus().toString());
    technicalInfosMap.put("lastIndexationDate", OffsetDateTime.now(ZoneOffset.UTC));
    technicalInfosMap.put("creationTime", publication.getCreationTime());
    technicalInfosMap.put("updateTime", publication.getUpdateTime());
    technicalInfosMap.put("lastStatusUpdate", publication.getLastStatusUpdate());
    List<StatusHistory> statusHistories = this.remotePublicationService.history(publication.getResId());
    this.putIfNotNull(technicalInfosMap, "firstValidation", this.getFirstValidationDateTime(statusHistories));
    publicationIndexEntry.getMetadata().put("technicalInfos", technicalInfosMap);

    return publicationIndexEntry;
  }

  private Map<String, Object> xmlMetadataToMap(Publication publication) {

    Object depositDoc = this.getDepositDoc(publication);
    DepositDocAdapter depositDocAdapter = this.getDepositDocAdapter(publication);

    Map<String, Object> properties = new HashMap<>();
    this.putIfNotNull(properties, "archiveId", publication.getArchiveId());
    this.putIfNotNull(properties, "type", depositDocAdapter.getType(depositDoc));
    this.putIfNotNull(properties, "subtype", depositDocAdapter.getSubtype(depositDoc));
    this.putIfNotNull(properties, "subsubtype", depositDocAdapter.getSubSubtype(depositDoc));
    this.putIfNotNull(properties, "languages", depositDocAdapter.getLanguages(depositDoc));

    this.putIfNotNull(properties, "doi", depositDocAdapter.getDOI(depositDoc));
    this.putIfNotNull(properties, "pmid", depositDocAdapter.getPmid(depositDoc));
    this.putIfNotNull(properties, "isbn", depositDocAdapter.getISBN(depositDoc));
    this.putIfNotNull(properties, "issn", depositDocAdapter.getISSN(depositDoc));
    this.putIfNotNull(properties, "arxiv", depositDocAdapter.getArxivId(depositDoc));
    this.putIfNotNull(properties, "mmsid", depositDocAdapter.getMmsid(depositDoc));
    this.putIfNotNull(properties, "repec", depositDocAdapter.getRepec(depositDoc));
    this.putIfNotNull(properties, "dblp", depositDocAdapter.getDblp(depositDoc));
    this.putIfNotNull(properties, "urn", depositDocAdapter.getURN(depositDoc));
    this.putIfNotNull(properties, "pmcid", depositDocAdapter.getPmcId(depositDoc));
    this.putIfNotNull(properties, "localNumber", depositDocAdapter.getLocalNumber(depositDoc));

    this.putIfNotNull(properties, "publisherVersionUrl", depositDocAdapter.getPublisherVersionUrl(depositDoc));
    this.putIfNotNull(properties, "note", depositDocAdapter.getNote(depositDoc));
    this.putIfNotNull(properties, "keywords", depositDocAdapter.getKeywords(depositDoc));
    this.putIfNotNull(properties, "discipline", depositDocAdapter.getDiscipline(depositDoc));
    this.putIfNotNull(properties, "mandator", depositDocAdapter.getMandator(depositDoc));
    this.putIfNotNull(properties, "publisherName", depositDocAdapter.getPublisherName(depositDoc));
    this.putIfNotNull(properties, "publisherPlace", depositDocAdapter.getPublisherPlace(depositDoc));
    this.putIfNotNull(properties, "edition", depositDocAdapter.getEdition(depositDoc));
    this.putIfNotNull(properties, "award", depositDocAdapter.getAward(depositDoc));
    this.putIfNotNull(properties, "isBeforeUnige", depositDocAdapter.isBeforeUnige(depositDoc));
    this.putIfNotNull(properties, "datasets", depositDocAdapter.getDatasets(depositDoc));
    this.putIfNotNull(properties, "aouCollection", depositDocAdapter.getAouCollection(depositDoc));

    this.fillTitle(depositDoc, depositDocAdapter, properties);
    this.fillOriginalTitle(depositDoc, depositDocAdapter, properties);
    this.fillContributors(depositDoc, depositDocAdapter, properties);
    this.fillAbstracts(depositDoc, depositDocAdapter, properties);
    this.fillDates(depositDoc, depositDocAdapter, properties);
    this.fillYears(depositDoc, depositDocAdapter, properties);
    this.fillPages(depositDoc, depositDocAdapter, properties);
    this.fillClassifications(depositDoc, depositDocAdapter, properties);
    this.fillCollections(depositDoc, depositDocAdapter, properties);
    this.fillFundings(depositDoc, depositDocAdapter, properties);
    this.fillStructures(depositDoc, depositDocAdapter, properties);
    this.fillResearchGroups(depositDoc, depositDocAdapter, properties);
    this.fillContainer(depositDoc, depositDocAdapter, properties);
    this.fillLinks(depositDoc, depositDocAdapter, properties);
    this.fillFiles(depositDoc, depositDocAdapter, properties);
    this.fillCorrections(depositDoc, depositDocAdapter, properties);

    this.fillFulltextVersions(depositDoc, depositDocAdapter, properties);
    this.fillEmbargoEndDates(depositDoc, depositDocAdapter, properties);
    this.fillLicenses(depositDoc, depositDocAdapter, properties);
    this.fillCurrentDiffusion(depositDoc, depositDocAdapter, properties);

    this.fillFulltexts(publication.getResId(), depositDoc, depositDocAdapter, properties);

    return properties;
  }

  private void fillTitle(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    this.putIfNotNull(properties, "title", depositDocAdapter.getTitleContent(depositDoc));
    this.putIfNotNull(properties, "titleLang", depositDocAdapter.getTitleLang(depositDoc));
  }

  private void fillOriginalTitle(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    this.putIfNotNull(properties, "originalTitle", depositDocAdapter.getOriginalTitleContent(depositDoc));
    this.putIfNotNull(properties, "originalTitleLang", depositDocAdapter.getOriginalTitleLang(depositDoc));
  }

  private void fillContributors(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> contributors = new ArrayList<>();
    Set<String> unigeDirectors = new HashSet<>();
    for (AbstractContributor abstractContributor : depositDocAdapter.getContributors(depositDoc)) {
      Map<String, String> contributor = new HashMap<>();
      if (abstractContributor instanceof ContributorDTO) {

        String firstName = ((ContributorDTO) abstractContributor).getFirstName();
        String lastName = ((ContributorDTO) abstractContributor).getLastName();
        String fullName = ((ContributorDTO) abstractContributor).getFullName();
        String cnIndividu = ((ContributorDTO) abstractContributor).getCnIndividu();
        String role = ((ContributorDTO) abstractContributor).getRole();

        contributor.put("type", "contributor");
        this.putIfNotNull(contributor, "firstName", firstName);
        this.putIfNotNull(contributor, "lastName", lastName);
        this.putIfNotNull(contributor, "fullName", fullName);
        this.putIfNotNull(contributor, "cnIndividu", cnIndividu);
        this.putIfNotNull(contributor, "role", role);
        this.putIfNotNull(contributor, "institution", ((ContributorDTO) abstractContributor).getInstitution());
        this.putIfNotNull(contributor, "email", ((ContributorDTO) abstractContributor).getEmail());
        this.putIfNotNull(contributor, "orcid", ((ContributorDTO) abstractContributor).getOrcid());
        this.putIfNotNull(contributor, "structure", ((ContributorDTO) abstractContributor).getStructure());

        if (!StringTool.isNullOrEmpty(cnIndividu) && AouConstants.CONTRIBUTOR_ROLE_DIRECTOR.equals(role)) {
          unigeDirectors.add(cnIndividu);
        }

      } else {
        this.putIfNotNull(contributor, "type", "collaboration");
        this.putIfNotNull(contributor, "name", ((CollaborationDTO) abstractContributor).getName());
      }

      if (!contributor.isEmpty()) {
        contributors.add(contributor);
      }
    }

    if (!contributors.isEmpty()) {
      properties.put("contributors", contributors);
    }

    if (!unigeDirectors.isEmpty()) {
      properties.put("unigeDirectors", unigeDirectors);
    }
  }

  private void fillAbstracts(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> abstracts = new ArrayList<>();
    for (MetadataTextLang metadataTextLang : depositDocAdapter.getAbstracts(depositDoc)) {
      Map<String, String> metadataMap = new HashMap<>();
      if (!StringTool.isNullOrEmpty(metadataTextLang.getContent())) {
        metadataMap.put("content", metadataTextLang.getContent());
      }
      if (!StringTool.isNullOrEmpty(metadataTextLang.getLang())) {
        metadataMap.put("language", metadataTextLang.getLang());
      }
      abstracts.add(metadataMap);
    }
    if (!abstracts.isEmpty()) {
      properties.put("abstracts", abstracts);
    }
  }

  private void fillDates(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> dates = new ArrayList<>();
    for (MetadataDates metadataDate : depositDocAdapter.getDates(depositDoc)) {
      Map<String, String> metadataMap = new HashMap<>();
      if (!StringTool.isNullOrEmpty(metadataDate.getContent())) {
        metadataMap.put("date", metadataDate.getContent());
      }
      if (!StringTool.isNullOrEmpty(metadataDate.getType())) {
        metadataMap.put("type", DateType.fromMetadataValue(metadataDate.getType()).name());
      }
      dates.add(metadataMap);
    }
    if (!dates.isEmpty()) {
      properties.put("dates", dates);
    }
  }

  private void fillYears(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    Set<Integer> years = new HashSet<>();
    for (MetadataDates metadataDate : depositDocAdapter.getDates(depositDoc)) {
      if (!StringTool.isNullOrEmpty(metadataDate.getContent())) {
        Integer year = this.getYear(metadataDate.getContent());
        if (year != null) {
          years.add(year);
        }
      }
    }
    if (!years.isEmpty()) {
      properties.put("years", years);
    }
  }

  private Integer getYear(String dateStr) {
    DateTimeFormatter formatter;
    for (String dateFormat : this.metadataDateFormat) {
      try {
        formatter = new DateTimeFormatterBuilder()
                .appendPattern(dateFormat)
                .parseDefaulting(ChronoField.DAY_OF_MONTH, 1)
                .parseDefaulting(ChronoField.MONTH_OF_YEAR, 1)
                .toFormatter();
        LocalDate date = LocalDate.parse(dateStr, formatter);
        return date.getYear();
      } catch (DateTimeParseException e) {
        continue;
      }
    }
    return null;
  }

  private void fillPages(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    Map<String, String> pages = new HashMap<>();
    this.putIfNotNull(pages, "paging", depositDocAdapter.getPaging(depositDoc));
    this.putIfNotNull(pages, "other", depositDocAdapter.getPagingOther(depositDoc));
    if (!pages.isEmpty()) {
      properties.put("pages", pages);
    }
  }

  private void fillClassifications(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> classifications = new ArrayList<>();
    for (MetadataClassification metadataClassification : depositDocAdapter.getClassifications(depositDoc)) {
      Map<String, String> classification = new HashMap<>();
      this.putIfNotNull(classification, "code", metadataClassification.getCode());
      this.putIfNotNull(classification, "item", metadataClassification.getItem());
      if (!classification.isEmpty()) {
        classifications.add(classification);
      }
    }
    if (!classifications.isEmpty()) {
      properties.put("classifications", classifications);
    }
  }

  private void fillCollections(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> collections = new ArrayList<>();
    for (MetadataCollection metadataCollection : depositDocAdapter.getCollections(depositDoc)) {
      Map<String, String> collection = new HashMap<>();
      this.putIfNotNull(collection, "name", metadataCollection.getName());
      this.putIfNotNull(collection, "number", metadataCollection.getNumber());
      if (!collection.isEmpty()) {
        collections.add(collection);
      }
    }
    if (!collections.isEmpty()) {
      properties.put("collections", collections);
    }
  }

  private void fillFundings(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> fundings = new ArrayList<>();
    for (MetadataFunding metadataFunding : depositDocAdapter.getFundings(depositDoc)) {
      Map<String, String> funding = new HashMap<>();
      this.putIfNotNull(funding, "name", metadataFunding.getName());
      this.putIfNotNull(funding, "code", metadataFunding.getCode());
      this.putIfNotNull(funding, "funder", metadataFunding.getFunder());
      this.putIfNotNull(funding, INDEX_ACRONYM_FIELD, metadataFunding.getAcronym());
      if (!funding.isEmpty()) {
        fundings.add(funding);
      }
    }
    if (!fundings.isEmpty()) {
      properties.put("fundings", fundings);
    }
  }

  private void fillStructures(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> structures = new ArrayList<>();
    Set<String> structuresWithParentsIds = new HashSet<>();
    for (Structure structure : depositDocAdapter.getStructures(depositDoc)) {
      Map<String, String> structureMap = new HashMap<>();
      this.putIfNotNull(structureMap, "name", structure.getName());
      this.putIfNotNull(structureMap, "code", structure.getCodeStruct());
      this.putIfNotNull(structureMap, "cnStruct", structure.getCnStructC());
      this.putIfNotNull(structureMap, INDEX_ACRONYM_FIELD, structure.getAcronym());
      this.putIfNotNull(structureMap, INDEX_RES_ID_FIELD, structure.getResId());
      if (!structureMap.isEmpty()) {
        structures.add(structureMap);
      }

      // fill parent structures ids
      structuresWithParentsIds.add(structure.getResId());
      structuresWithParentsIds.addAll(this.remoteStructureService.getParentStructureIds(structure.getResId()));
    }
    if (!structures.isEmpty()) {
      properties.put("structures", structures);
    }
    if (!structuresWithParentsIds.isEmpty()) {
      properties.put("structuresWithParentsIds", structuresWithParentsIds);
    }
  }

  private void fillResearchGroups(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> researchGroups = new ArrayList<>();
    for (ResearchGroup researchGroup : depositDocAdapter.getResearchGroups(depositDoc)) {
      Map<String, String> researchGroupMap = new HashMap<>();
      this.putIfNotNull(researchGroupMap, "name", researchGroup.getName());
      this.putIfNotNull(researchGroupMap, "code", researchGroup.getCode());
      this.putIfNotNull(researchGroupMap, INDEX_ACRONYM_FIELD, researchGroup.getAcronym());
      this.putIfNotNull(researchGroupMap, INDEX_RES_ID_FIELD, researchGroup.getResId());
      if (!researchGroupMap.isEmpty()) {
        researchGroups.add(researchGroupMap);
      }
    }
    if (!researchGroups.isEmpty()) {
      properties.put("researchGroups", researchGroups);
    }
  }

  private void fillContainer(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    Map<String, String> containerMap = new HashMap<>();
    this.putIfNotNull(containerMap, "title", depositDocAdapter.getContainerTitle(depositDoc));
    this.putIfNotNull(containerMap, "volume", depositDocAdapter.getContainerVolume(depositDoc));
    this.putIfNotNull(containerMap, "issue", depositDocAdapter.getContainerIssue(depositDoc));
    this.putIfNotNull(containerMap, "specialIssue", depositDocAdapter.getContainerSpecialIssue(depositDoc));
    this.putIfNotNull(containerMap, "editor", depositDocAdapter.getContainerEditor(depositDoc));
    this.putIfNotNull(containerMap, "conferencePlace", depositDocAdapter.getContainerConferencePlace(depositDoc));
    this.putIfNotNull(containerMap, "conferenceDate", depositDocAdapter.getContainerConferenceDate(depositDoc));
    this.putIfNotNull(containerMap, "conferenceSubtitle", depositDocAdapter.getContainerConferenceSubtitle(depositDoc));
    if (!containerMap.isEmpty()) {
      properties.put("container", containerMap);
    }
  }

  private void fillLinks(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> links = new ArrayList<>();
    for (MetadataLink metadataLink : depositDocAdapter.getLinks(depositDoc)) {
      Map<String, String> link = new HashMap<>();
      this.putIfNotNull(link, "type", metadataLink.getType());
      this.putIfNotNull(link, "target", metadataLink.getTarget());
      if (!link.isEmpty()) {
        links.add(link);
      }
    }
    if (!links.isEmpty()) {
      properties.put("links", links);
    }
  }

  private void fillCorrections(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, String>> corrections = new ArrayList<>();
    for (MetadataCorrection metadataCorrection : depositDocAdapter.getCorrections(depositDoc)) {
      Map<String, String> correction = new HashMap<>();
      this.putIfNotNull(correction, "note", metadataCorrection.getNote());
      this.putIfNotNull(correction, "doi", metadataCorrection.getDoi());
      this.putIfNotNull(correction, "pmid", metadataCorrection.getPmid());
      if (!correction.isEmpty()) {
        corrections.add(correction);
      }
    }
    if (!corrections.isEmpty()) {
      properties.put("corrections", corrections);
    }
  }

  private void fillFiles(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<Map<String, Object>> files = new ArrayList<>();
    for (MetadataFile metadataFile : depositDocAdapter.getFiles(depositDoc)) {
      if (!metadataFile.getType().equals(AouConstants.DOCUMENT_FILE_TYPE_VIGNETTE)) {
        Map<String, Object> fileMap = this.getFileMap(metadataFile);
        this.fillFileTypeLevel(depositDoc, depositDocAdapter, metadataFile, fileMap);
        
        if (!fileMap.isEmpty()) {
          files.add(fileMap);
        }
      } else {
        this.fillThumbnail(metadataFile, properties);
      }
    }
    if (!files.isEmpty()) {
      properties.put("files", files);
    }
  }

  private void fillFileTypeLevel(Object depositDoc, DepositDocAdapter depositDocAdapter, MetadataFile metadataFile,
          Map<String, Object> fileMap) {
    String publicationSubtypeId = this.getPublicationSubtypeIdByName(depositDocAdapter.getSubtype(depositDoc));
    String documentFileTypeId = this.getDocumentFileTypeIdByName(metadataFile.getType());

    List<PublicationSubtypeDocumentFileTypeDTO> pubSubDocTypes = this.remotePublicationSubtypeService.findAllPublicationSubtypeDocumentFileTypesDTO(
            publicationSubtypeId, PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE, Sort.by("sortValue"))).getData();
    Optional<PublicationSubtypeDocumentFileTypeDTO> psdtOpt = pubSubDocTypes.stream()
            .filter(psdt -> psdt.getDocumentFileTypeId().equals(documentFileTypeId) && psdt.getPublicationSubtypeId()
                    .equals(publicationSubtypeId)).findFirst();
    if (psdtOpt.isPresent()) {
      this.putIfNotNull(fileMap, "typeLevel", psdtOpt.get().getLevel().name());
    }
  }

  private String getPublicationSubtypeIdByName(String publicationSubtypeName) {
    List<PublicationSubtypeDTO> publicationSubtypes = this.remotePublicationSubtypeService.findAllDTO(
            PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE, Sort.by("sortValue"))).getData();

    Optional<PublicationSubtypeDTO> publicationSubtypeOpt = publicationSubtypes.stream()
            .filter(ps -> ps.getName().equals(publicationSubtypeName)).findFirst();
    if (publicationSubtypeOpt.isPresent()) {
      return publicationSubtypeOpt.get().getResId();
    } else {
      throw new SolidifyRuntimeException("PublicationSubytpe with name '" + publicationSubtypeName + "' not found");
    }
  }

  private String getDocumentFileTypeIdByName(String fileTypeName) {
    List<DocumentFileType> documentFileTypes = this.remoteDocumentFileTypeService.findAllWithCache(
            PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE, Sort.by("value"))).getData();

    Optional<DocumentFileType> documentFileTypeOpt = documentFileTypes.stream().filter(dft -> dft.getValue().equals(fileTypeName)).findFirst();
    if (documentFileTypeOpt.isPresent()) {
      return documentFileTypeOpt.get().getResId();
    } else {
      throw new SolidifyRuntimeException("DocumentFileType with name '" + fileTypeName + "' not found");
    }
  }

  private void fillThumbnail(MetadataFile metadataFile, Map<String, Object> properties) {
    Map<String, Object> thumbnailMap = this.getFileMap(metadataFile);
    if (!thumbnailMap.isEmpty()) {
      properties.put("thumbnail", thumbnailMap);
    }
  }

  private Map<String, Object> getFileMap(MetadataFile metadataFile) {
    Map<String, Object> file = new HashMap<>();
    this.putIfNotNull(file, INDEX_RES_ID_FIELD, metadataFile.getResId());
    this.putIfNotNull(file, "type", metadataFile.getType());
    this.putIfNotNull(file, "name", metadataFile.getName());
    this.putIfNotNull(file, "size", metadataFile.getSize());
    this.putIfNotNull(file, "mimeType", metadataFile.getMimeType());
    this.putIfNotNull(file, "license", metadataFile.getLicense());
    this.putIfNotNull(file, "label", metadataFile.getLabel());

    DocumentFile.AccessLevel currentAccessLevel = DocumentFile.AccessLevel.fromMetadataValue(metadataFile.getAccessLevel());

    MetadataEmbargo embargo = metadataFile.getEmbargo();
    if (embargo != null) {
      this.putIfNotNull(file, "finalAccessLevel", DocumentFile.AccessLevel.fromMetadataValue(metadataFile.getAccessLevel()).name());
      this.putIfNotNull(file, "embargoAccessLevel", DocumentFile.AccessLevel.fromMetadataValue(embargo.getAccessLevel()).name());
      this.putIfNotNull(file, "embargoEndDate", embargo.getEndDate());

      // if embargo is not over, check current access level
      if (embargo.getEndDate().isAfter(LocalDate.now())) {
        currentAccessLevel = DocumentFile.AccessLevel.fromMetadataValue(embargo.getAccessLevel());
      }
    }

    this.putIfNotNull(file, "accessLevel", currentAccessLevel.name());

    return file;
  }

  /*****************************************************************************************************/

  private void fillFulltextVersions(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {

    Set<String> fulltextVersions = new HashSet<>();

    for (MetadataFile metadataFile : depositDocAdapter.getFiles(depositDoc)) {
      String fulltextVersion = this.getFulltextVersion(metadataFile);
      if (!StringTool.isNullOrEmpty(fulltextVersion)) {
        fulltextVersions.add(fulltextVersion);
      }
    }

    this.putIfNotNull(properties, "fulltextVersions", fulltextVersions);
  }

  private String getFulltextVersion(MetadataFile metadataFile) {
    String type = metadataFile.getType();
    if (type.contains(AouConstants.DOCUMENT_PUBLISHED_VERSION_EN)) {
      return AouConstants.PUBLISHED_VERSION_TEXT;
    } else if (type.contains(AouConstants.DOCUMENT_ACCEPTED_VERSION_EN)) {
      return AouConstants.ACCEPTED_VERSION_TEXT;
    } else if (type.contains(AouConstants.DOCUMENT_SUBMITTED_VERSION_EN)) {
      return AouConstants.SUBMITTED_VERSION_TEXT;
    } else {
      switch (type) {
        case AouConstants.DOCUMENT_FILE_TYPE_ADDENDUM:
        case AouConstants.DOCUMENT_FILE_TYPE_CORRIGENDUM:
        case AouConstants.DOCUMENT_FILE_TYPE_ERRATUM:
        case AouConstants.DOCUMENT_FILE_TYPE_RETRACTION:
          return AouConstants.PUBLISHED_VERSION_TEXT;

        case AouConstants.DOCUMENT_FILE_TYPE_MASTER_THESIS:
        case AouConstants.DOCUMENT_FILE_TYPE_POSTER:
        case AouConstants.DOCUMENT_FILE_TYPE_THESIS:
        case AouConstants.DOCUMENT_FILE_TYPE_PRESENTATION:
          return AouConstants.ACCEPTED_VERSION_TEXT;

        case AouConstants.DOCUMENT_FILE_TYPE_REPORT:
          return AouConstants.UNKNOWN_TEXT;

        case AouConstants.DOCUMENT_FILE_TYPE_PREPRINT:
          return AouConstants.DRAFT_TEXT;

        case AouConstants.DOCUMENT_FILE_TYPE_EXTRACT:
          return AouConstants.EXTRACT_TEXT;

        case AouConstants.DOCUMENT_FILE_TYPE_VIGNETTE:
        default:
          return null;
      }
    }
  }

  private void fillEmbargoEndDates(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {

    Set<LocalDate> embargoEndDates = new HashSet<>();

    for (MetadataFile metadataFile : depositDocAdapter.getFiles(depositDoc)) {
      if (metadataFile.getEmbargo() != null) {
        embargoEndDates.add(metadataFile.getEmbargo().getEndDate());
      }
    }

    this.putIfNotNull(properties, "embargoEndDates", embargoEndDates);
  }

  private void fillLicenses(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    Set<String> licenses = new HashSet<>();
    for (MetadataFile metadataFile : depositDocAdapter.getFiles(depositDoc)) {
      if (!StringTool.isNullOrEmpty(metadataFile.getLicense())) {
        licenses.add(metadataFile.getLicense());
      }
    }
    this.putIfNotNull(properties, "licenses", licenses);
  }

  private void fillCurrentDiffusion(Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {

    Map<String, List<String>> accessLevelByTypes = new HashMap<>();

    // compute a Map with a list of access levels for each type of file
    for (MetadataFile metadataFile : depositDocAdapter.getFiles(depositDoc)) {
      String versionText = this.getFulltextVersion(metadataFile);
      String currentAccessLevel = this.getCurrentDiffusion(metadataFile);
      if (!StringTool.isNullOrEmpty(versionText) && !StringTool.isNullOrEmpty(currentAccessLevel)) {
        accessLevelByTypes.computeIfAbsent(versionText, x -> new ArrayList<>()).add(currentAccessLevel);
      }
    }

    // according to this map, determine the global access level of the deposit

    Set<String> accessLevelsForPublishedAcceptedSubmittedTypes = new HashSet<>();
    accessLevelByTypes.entrySet().stream()
            .filter(m -> m.getKey().equals(AouConstants.PUBLISHED_VERSION_TEXT)
                    || m.getKey().equals(AouConstants.ACCEPTED_VERSION_TEXT)
                    || m.getKey().equals(AouConstants.SUBMITTED_VERSION_TEXT))
            .forEach(m -> accessLevelsForPublishedAcceptedSubmittedTypes.addAll(m.getValue()));
    String diffusion = this.getMoreOpenAccessLevel(accessLevelsForPublishedAcceptedSubmittedTypes);

    if (StringTool.isNullOrEmpty(diffusion)) {
      Set<String> accessLevelsForUnknownDraftExtractTypes = new HashSet<>();
      accessLevelByTypes.entrySet().stream()
              .filter(m -> m.getKey().equals(AouConstants.UNKNOWN_TEXT)
                      || m.getKey().equals(AouConstants.DRAFT_TEXT)
                      || m.getKey().equals(AouConstants.EXTRACT_TEXT))
              .forEach(m -> accessLevelsForUnknownDraftExtractTypes.addAll(m.getValue()));
      diffusion = this.getMoreOpenAccessLevel(accessLevelsForUnknownDraftExtractTypes);
    }

    this.putIfNotNull(properties, "diffusion", diffusion);
    this.putIfNotNull(properties, "accessLevelByTypes", accessLevelByTypes);
  }

  private String getMoreOpenAccessLevel(Set<String> accessLevels) {
    if (accessLevels.contains(DocumentFile.AccessLevel.PUBLIC.metadataValue())) {
      return AouConstants.OPEN_ACCESS_TEXT;
    } else if (accessLevels.contains(AouConstants.EMBARGOED_ACCESS_TEXT)) {
      return AouConstants.EMBARGOED_ACCESS_TEXT;
    } else if (accessLevels.contains(DocumentFile.AccessLevel.RESTRICTED.metadataValue())) {
      return AouConstants.RESTRICTED_ACCESS_TEXT;
    } else if (accessLevels.contains(DocumentFile.AccessLevel.PRIVATE.metadataValue())) {
      return AouConstants.CLOSED_ACCESS_TEXT;
    }
    return null;
  }

  private String getCurrentDiffusion(MetadataFile metadataFile) {
    if (metadataFile.getEmbargo() != null && metadataFile.getEmbargo().getEndDate().isAfter(LocalDate.now())) {
      return AouConstants.EMBARGOED_ACCESS_TEXT;
    } else {
      return metadataFile.getAccessLevel();
    }
  }

  private OffsetDateTime getFirstValidationDateTime(List<StatusHistory> statusHistories) {
    for (StatusHistory statusHistory : statusHistories) {
      if (statusHistory.getStatus().equals(Publication.PublicationStatus.COMPLETED.toString())) {
        return statusHistory.getChangeTime();
      }
    }
    return null;
  }

  /*****************************************************************************************************/

  private void fillFulltexts(String publicationId, Object depositDoc, DepositDocAdapter depositDocAdapter, Map<String, Object> properties) {
    List<String> fulltexts = new ArrayList<>();
    for (MetadataFile metadataFile : depositDocAdapter.getFiles(depositDoc)) {
      String currentAccessLevelText = this.getCurrentDiffusion(metadataFile);
      if (!currentAccessLevelText.equals(AouConstants.EMBARGOED_ACCESS_TEXT)
              && DocumentFile.AccessLevel.fromMetadataValue(currentAccessLevelText) == DocumentFile.AccessLevel.PUBLIC
              && !isFileTypeModeDiffusionOrAgreement(metadataFile.getType())) {
        // Index only fulltext of files that are currently public
        String fulltext = null;
        if (!StringTool.isNullOrEmpty(metadataFile.getResId())) {
          fulltext = this.remoteDocumentFileService.fulltext(metadataFile.getResId());
        } else {
          // previous metadata version didn't contain resId
          Map<String, String> queryParams = new HashMap<>();
          queryParams.put("size", metadataFile.getSize().toString());
          queryParams.put("fileName", metadataFile.getName());
          List<DocumentFile> documentFiles = this.remoteDocumentFileService.listByPublication(publicationId, queryParams);
          if (documentFiles.size() == 1) {
            fulltext = this.remoteDocumentFileService.fulltext(documentFiles.get(0).getResId());
          }
        }
        if (!StringTool.isNullOrEmpty(fulltext)) {
          fulltexts.add(fulltext);
        }
      }
    }
    if (!fulltexts.isEmpty()) {
      properties.put("fulltexts", fulltexts);
    }
  }

  private boolean isFileTypeModeDiffusionOrAgreement(String type) {
    return type.equals(DOCUMENT_FILE_TYPE_PUBLICATION_MODE_VALUE) || type.equals(DOCUMENT_FILE_TYPE_AGREEMENT_VALUE);
  }

  private void putIfNotNull(Map map, String propertyName, Object value) {
    if (value != null && !StringTool.isNullOrEmpty(value.toString())) {
      map.put(propertyName, value);
    }
  }

  /*****************************************************************************************************/

  public Object getOaiDcMetadata(PublicationMetadata archive) {
    // TODO Auto-generated method stub
    return null;
  }

  public PublicationMetadata getMetatadata(String resIdFromOaiId) {
    // TODO Auto-generated method stub
    return null;
  }
}
