package ch.unige.aou.service;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.event.EventListener;
import org.springframework.jms.core.JmsTemplate;
import org.springframework.transaction.event.TransactionalEventListener;

import ch.unige.solidify.message.CacheMessage;
import ch.unige.solidify.message.ResourceCacheMessage;
import ch.unige.solidify.model.index.IndexFieldAlias;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.config.AouProperties;
import ch.unige.aou.message.EmailMessage;
import ch.unige.aou.message.PublicationFromFedoraMessage;
import ch.unige.aou.message.PublicationIndexingMessage;
import ch.unige.aou.message.PublicationMessage;
import ch.unige.aou.message.PublicationMetadataUpgradeMessage;
import ch.unige.aou.model.publication.DocumentFileType;
import ch.unige.aou.model.publication.PublicationSubtype;
import ch.unige.aou.model.settings.Structure;

public abstract class JmsService extends AouService {
  private static final Logger log = LoggerFactory.getLogger(JmsService.class);

  @Autowired
  private JmsTemplate queueJmsTemplate;

  @Autowired
  @Qualifier("topicJmsTemplate")
  private JmsTemplate topicJmsTemplate;

  private String publicationsQueue;
  private String publicationsIndexingQueue;
  private String publicationsMetadataUpgradeQueue;
  private String emailsQueue;
  private String publicationsFromFedoraQueue;
  private final String cacheTopic;

  public JmsService(AouProperties aouProperties, MessageService messageService) {
    super(messageService);

    this.publicationsQueue = aouProperties.getQueue().getPublications();
    this.publicationsIndexingQueue = aouProperties.getQueue().getPublicationsIndexing();
    this.publicationsMetadataUpgradeQueue = aouProperties.getQueue().getPublicationsMetadataUpgrade();
    this.emailsQueue = aouProperties.getQueue().getEmails();
    this.publicationsFromFedoraQueue = aouProperties.getQueue().getPublicationsFromFedora();
    this.cacheTopic = aouProperties.getTopic().getCache();
  }

  protected void convertAndSendToQueue(String queue, Object event) {
    if (log.isTraceEnabled()) {
      log.trace("Sending message to queue {} => {}", queue, event);
    }
    this.queueJmsTemplate.convertAndSend(queue, event);
  }

  protected void convertAndSendToTopic(String topic, Object event) {
    if (log.isTraceEnabled()) {
      log.trace("Sending message to topic {} => {}", topic, event);
    }
    this.topicJmsTemplate.convertAndSend(topic, event);
  }

  @EventListener
  private void sendMessage(PublicationMessage publicationMessage) {
    this.convertAndSendToQueue(this.publicationsQueue, publicationMessage);
  }

  @EventListener
  private void sendMessage(PublicationMetadataUpgradeMessage publicationMetadataUpgradeMessage) {
    this.convertAndSendToQueue(this.publicationsMetadataUpgradeQueue, publicationMetadataUpgradeMessage);
  }

  @EventListener
  private void sendMessage(PublicationIndexingMessage publicationIndexingMessage) {
    this.convertAndSendToQueue(this.publicationsIndexingQueue, publicationIndexingMessage);
  }

  @EventListener
  private void sendMessage(EmailMessage emailMessage) {
    this.convertAndSendToQueue(this.emailsQueue, emailMessage);
  }

  @EventListener
  private void sendMessage(PublicationFromFedoraMessage publicationFromFedoraMessage) {
    this.convertAndSendToQueue(this.publicationsFromFedoraQueue, publicationFromFedoraMessage);
  }

  @TransactionalEventListener(fallbackExecution = true)
  private void sendMessage(CacheMessage cacheMessage) {
    // Only send JMS messages for resources that are effectively put in cache
    if (cacheMessage instanceof ResourceCacheMessage &&
            (((ResourceCacheMessage) cacheMessage).getResourceClass() == IndexFieldAlias.class
                    || ((ResourceCacheMessage) cacheMessage).getResourceClass() == Structure.class
                    || ((ResourceCacheMessage) cacheMessage).getResourceClass() == PublicationSubtype.class
                    || ((ResourceCacheMessage) cacheMessage).getResourceClass() == DocumentFileType.class)) {
      this.convertAndSendToTopic(this.cacheTopic, cacheMessage);
    }
  }
}
